from __future__ import annotations

import random
from typing import Iterable

from negmas.outcomes import Outcome

from .base_ufun import BaseUtilityFunction
from .protocols import InverseUFun

__all__ = [
    "PresortingInverseUtilityFunction",
    "SamplingInverseUtilityFunction",
]


class SamplingInverseUtilityFunction(InverseUFun):
    """
    A utility function inverter that uses sampling.

    Nothing is done during initialization so the fixed cost of this inverter is minimal.
    Nevertheless, each time the system is asked to find an outcome within some range, it uses
    random sampling which is very inefficient and suffers from the curse of dimensinality.
    """

    def __init__(self, ufun: BaseUtilityFunction, max_samples_per_call: int = 10_000):
        self._ufun = ufun
        self.max_samples_per_call = max_samples_per_call
        self._initialized = True

    @property
    def initialized(self):
        return self._initialized

    @property
    def ufun(self):
        return self._ufun

    def init(self):
        pass

    def all(
        self,
        rng: float | tuple[float, float],
    ) -> list[Outcome]:
        """
        Finds all outcomes with in the given utility value range

        Args:
            rng: The range. If a value, outcome utilities must match it exactly

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        raise ValueError(
            f"Cannot find all outcomes in a range using a SamplingInverseUtilityFunction. Try a PresortedInverseUtilityFunction"
        )

    def some(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
        n: int | None = None,
    ) -> list[Outcome]:
        """
        Finds some outcomes with the given utility value (if discrete, all)

        Args:
            rng: The range. If a value, outcome utilities must match it exactly
            n: The maximum number of outcomes to return

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        if not isinstance(rng, Iterable):
            rng = (rng - 1e-5, rng + 1e-5)
        if not n:
            n = self.max_samples_per_call
        else:
            n *= 3
        if not self._ufun.outcome_space:
            return []
        outcomes = list(self._ufun.outcome_space.sample(n, False, False))
        mn, mx = rng
        u = self.ufun.eval_normalized if normalized else self.ufun.eval
        return [_ for _ in outcomes if mn <= u(_) <= mx]

    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        some = self.some(rng, normalized)
        if not isinstance(rng, Iterable):
            rng = (rng, rng)
        worst_util, worst = float("inf"), None
        for o in some:
            util = self._ufun(o)
            if util < worst_util:
                worst_util, worst = util, o
        return worst

    def best_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        some = self.some(rng, normalized)
        if not isinstance(rng, Iterable):
            rng = (rng, rng)
        best_util, best = float("-inf"), None
        for o in some:
            util = self._ufun(o)
            if util < best_util:
                best_util, best = util, o
        return best

    def one_in(
        self, rng: float | tuple[float, float], normalized: float
    ) -> Outcome | None:
        if not self._ufun.outcome_space:
            return None
        if not isinstance(rng, Iterable):
            rng = (rng - 1e-5, rng + 1e-5)
        u = self.ufun.eval_normalized if normalized else self.ufun.eval
        for _ in range(self.max_samples_per_call):
            o = list(self._ufun.outcome_space.sample(1))[0]
            if rng[0] <= u(o) <= rng[1]:
                return o
        return None

    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:
            These may be different from the results of `ufun.minmax()` as they can be approximate.
        """
        return self.min(), self.max()

    def extreme_outcomes(self) -> tuple[Outcome, Outcome]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """
        return self.worst(), self.best()

    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """
        return self.one_in(rng, normalized)


class PresortingInverseUtilityFunction(InverseUFun):
    """
    A utility function inverter that uses pre-sorting.

    The outcome-space is sampled if it is continuous and enumerated if it is discrete
    during the call to `init()` and an ordered list of outcomes with their utility
    values is then cached.
    """

    def __init__(
        self, ufun: BaseUtilityFunction, levels: int = 10, max_cache_size: int = 100_000
    ):
        self._ufun = ufun
        self.max_cache_size = max_cache_size
        self.levels = levels
        self._initialized = False
        self._ordered_outcomes: list[tuple[float, Outcome]] = []

    @property
    def initialized(self):
        return self._initialized

    @property
    def ufun(self):
        return self._ufun

    def reset(self) -> None:
        self._initialized = False
        self._orddered_outcomes = []

    def init(self):
        outcome_space = self._ufun.outcome_space
        if outcome_space is None:
            raise ValueError("Cannot find the outcome space.")
        self._worst, self._best = self._ufun.extreme_outcomes()
        self._min, self._max = float(self._ufun(self._worst)), float(
            self._ufun(self._best)
        )
        self._range = self._max - self._min
        self._offset = self._min / self._range if self._range > 1e-5 else self._min
        for l in range(self.levels, 0, -1):
            n = outcome_space.cardinality_if_discretized(l)
            if n <= self.max_cache_size:
                break
        else:
            raise ValueError(
                f"Cannot discretize keeping cach size at {self.max_cache_size}. Outocme space cardinality is {outcome_space.cardinality}\nOutcome space: {outcome_space}"
            )
        os = outcome_space.to_discrete(levels=l, max_cardinality=self.max_cache_size)
        if os.cardinality <= self.max_cache_size:
            outcomes = list(os.sample(self.max_cache_size, False, False))
        else:
            outcomes = list(os.enumerate())[: self.max_cache_size]
        utils = [
            float(self._ufun.eval_normalized(_, above_reserve=False)) for _ in outcomes
        ]
        self._ordered_outcomes = sorted(zip(utils, outcomes), key=lambda x: -x[0])
        self._initialized = True

    def _normalize_range(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
    ) -> tuple[float, float]:
        if not isinstance(rng, Iterable):
            rng = (rng - 1e-5, rng + 1e-5)
        if normalized:
            return rng
        mn, mx = self._min, self._max
        d = mx - mn
        if d < 1e-8:
            return tuple(1.0 for _ in rng)
        return tuple((_ - mn) / d for _ in rng)

    def some(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
        n: int | None = None,
    ) -> list[Outcome]:
        """
        Finds some outcomes with the given utility value (if discrete, all)

        Args:
            rng: The range. If a value, outcome utilities must match it exactly
            n: The maximum number of outcomes to return

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized)
        mn, mx = rng
        # todo use bisection
        results = []
        for util, w in self._ordered_outcomes:
            if util > mx:
                continue
            if util < mn:
                break
            results.append(w)
        if n and len(results) >= n:
            return random.sample(results, n)
        return results

    def all(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
    ) -> list[Outcome]:
        """
        Finds all outcomes with in the given utility value range

        Args:
            rng: The range. If a value, outcome utilities must match it exactly

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        os_ = self._ufun.outcome_space
        if not os_:
            raise ValueError(f"Unkonwn outcome space. Cannot invert the ufun")

        if os_.is_discrete():
            return self.some(rng, normalized)
        raise ValueError(
            f"Cannot find all outcomes in a range for a continous outcome space (there is in general an infinite number of them)"
        )

    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized)
        mn, mx = rng
        if not self._ordered_outcomes:
            return None
        wbefore = self._ordered_outcomes[0][1]
        for i, (util, w) in enumerate(self._ordered_outcomes):
            if util > mx:
                continue
            if util < mn:
                return wbefore
            wbefore = w
        return wbefore

    def best_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized)
        mn, mx = rng
        for util, w in self._ordered_outcomes:
            if util > mx:
                continue
            if util < mn:
                return None
            return w
        return None

    def one_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        lst = self.some(rng, normalized)
        if not lst:
            return None
        return lst[random.randint(0, len(lst) - 1)]

    def within_fractions(self, rng: tuple[float, float]) -> list[Outcome]:
        """
        Finds outocmes within the given fractions of utility values (the fractions must be between zero and one)
        """
        if not self._ufun.is_stationary():
            self.init()
        n = len(self._ordered_outcomes)
        rng = (max(rng[0] * n, 0), min(rng[1] * n, len(self._ordered_outcomes)))
        return [_[1] for _ in self._ordered_outcomes[int(rng[0]) : int(rng[1])]]

    def within_indices(self, rng: tuple[int, int]) -> list[Outcome]:
        """
        Finds outocmes within the given indices with the best at index 0 and the worst at largest index.

        Remarks:
            - Works only for discrete outcome spaces
        """
        if not self._ufun.is_stationary():
            self.init()
        rng = (max(rng[0], 0), min(rng[1], len(self._ordered_outcomes)))
        return [_[1] for _ in self._ordered_outcomes[rng[0] : rng[1]]]

    def min(self) -> float:
        """
        Finds the minimum utility value
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError(f"No outcomes to find the best")
        return self._ordered_outcomes[-1][0]

    def max(self) -> float:
        """
        Finds the maximum utility value
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError(f"No outcomes to find the best")
        return self._ordered_outcomes[0][0]

    def worst(self) -> Outcome:
        """
        Finds the worst  outcome
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError(f"No outcomes to find the best")
        return self._ordered_outcomes[-1][1]

    def best(self) -> Outcome:
        """
        Finds the best  outcome
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError(f"No outcomes to find the best")
        return self._ordered_outcomes[0][1]

    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:
            These may be different from the results of `ufun.minmax()` as they can be approximate.
        """
        return self.min(), self.max()

    def extreme_outcomes(self) -> tuple[Outcome, Outcome]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """
        return self.worst(), self.best()

    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """
        return self.one_in(rng, normalized)



import warnings

__all__ = [
    "warn",
    "deprecated",
    "NegmasWarning",
    "NegmasIgnoredValueWarning",
    "NegmasBridgePathWarning",
    "NegmasBridgeProcessWarning",
    "NegmasInfiniteNegotiationWarning",
    "NegmasStepAndTimeLimitWarning",
    "NegmasBrdigeParsingWarning",
    "NegmasSarializationWarning",
    "NegmasIOWarning",
    "NegmasShutdownWarning",
    "NegmasWorldConfigWarning",
    "NegmasImportWarning",
    "NegmasUnusedValueWarning",
    "NegmasMemoryWarning",
    "NegmasSpeedWarning",
    "NegmasCaughtExceptionWarning",
    "NegmasVisualizationWarning",
    "NegmasNoResponseWarning",
    "NegmasLoggingWarning",
    "NegmasNumericWarning",
    "NegmasDoubleAssignmentWarning",
    "NegmasUnexpectedValueWarning",
]


class NegmasWarning(UserWarning):
    ...


def warn(message, category=NegmasWarning, stacklevel=2, source=None):
    """Issues a warning to the user. Defaults to `NegmasWarning` and stacklevel of 2."""
    return warnings.warn(message, category, stacklevel, source)


def deprecated(message):
    """Issues a deprecation warning"""
    return warnings.warn(
        message, category=DeprecationWarning, stacklevel=2, source=None
    )


class NegmasBridgePathWarning(NegmasWarning):
    ...


class NegmasBridgeProcessWarning(NegmasWarning):
    ...


class NegmasInfiniteNegotiationWarning(NegmasWarning):
    ...


class NegmasStepAndTimeLimitWarning(NegmasWarning):
    ...


class NegmasBrdigeParsingWarning(NegmasWarning):
    ...


class NegmasSarializationWarning(NegmasWarning):
    ...


class NegmasIOWarning(NegmasWarning):
    ...


class NegmasShutdownWarning(NegmasWarning):
    ...


class NegmasWorldConfigWarning(NegmasWarning):
    ...


class NegmasImportWarning(NegmasWarning, ImportWarning):
    ...


class NegmasUnusedValueWarning(NegmasWarning):
    ...


class NegmasMemoryWarning(NegmasWarning):
    ...


class NegmasCaughtExceptionWarning(NegmasWarning):
    ...


class NegmasVisualizationWarning(NegmasWarning):
    ...


class NegmasNoResponseWarning(NegmasWarning):
    ...


class NegmasLoggingWarning(NegmasWarning):
    ...


class NegmasNumericWarning(NegmasWarning):
    ...


class NegmasSpeedWarning(NegmasWarning):
    ...


class NegmasIgnoredValueWarning(NegmasWarning):
    ...


class NegmasDoubleAssignmentWarning(NegmasWarning):
    ...


class NegmasUnexpectedValueWarning(NegmasWarning):
    ...



from __future__ import annotations

from click.testing import CliRunner

from negmas.scripts.app import cli as main


def test_main():
    runner = CliRunner()
    result = runner.invoke(main, [])
    assert len(result.output) > 0
    assert result.exit_code == 0



"""
Defines import/export functionality
"""
from __future__ import annotations

import xml.etree.ElementTree as ET
from os import PathLike, listdir
from pathlib import Path
from typing import Any, Callable, Iterable, Sequence

from attr import define

from negmas.helpers.inout import dump
from negmas.outcomes.outcome_space import make_os
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, serialize

from .mechanisms import Mechanism
from .negotiators import Negotiator
from .outcomes import CartesianOutcomeSpace, Issue, issues_from_genius
from .preferences import (
    DiscountedUtilityFunction,
    UtilityFunction,
    conflict_level,
    make_discounted_ufun,
    nash_point,
    opposition_level,
    pareto_frontier,
    winwin_level,
)
from .preferences.value_fun import TableFun
from .sao import SAOMechanism

__all__ = [
    "Scenario",
    "load_genius_domain",
    "load_genius_domain_from_folder",
    "find_domain_and_utility_files",
    "get_domain_issues",
]

STATS_MAX_CARDINALITY = 100_1000


@define
class Scenario:
    """
    A class representing a negotiation domain
    """

    agenda: CartesianOutcomeSpace
    ufuns: tuple[UtilityFunction, ...]
    mechanism_type: type[Mechanism] | None
    mechanism_params: dict

    @property
    def issues(self) -> tuple[Issue, ...]:
        return self.agenda.issues

    def to_genius_files(self, domain_path: Path, ufun_paths: list[Path]):
        """
        Save domain and ufun files to the `path` as XML.
        """
        domain_path = Path(domain_path)
        ufun_paths = [Path(_) for _ in ufun_paths]
        if len(self.ufuns) != len(ufun_paths):
            raise ValueError(
                f"I have {len(self.ufuns)} ufuns but {len(ufun_paths)} paths were passed!!"
            )
        domain_path.parent.mkdir(parents=True, exist_ok=True)
        self.agenda.to_genius(domain_path)
        for ufun, path in zip(self.ufuns, ufun_paths):
            ufun.to_genius(path, issues=self.issues)
        return self

    def to_genius_folder(self, path: Path):
        """
        Save domain and ufun files to the `path` as XML.
        """
        path.mkdir(parents=True, exist_ok=True)
        domain_name = self.agenda.name.split("/")[-1] if self.agenda.name else "domain"
        ufun_names = [_.name.split("/")[-1] for _ in self.ufuns]
        self.agenda.to_genius(path / domain_name)
        for ufun, name in zip(self.ufuns, ufun_names):
            ufun.to_genius(path / name, issues=self.issues)
        return self

    @property
    def n_negotiators(self) -> int:
        return len(self.ufuns)

    @property
    def n_issues(self) -> int:
        return len(self.agenda.issues)

    @property
    def issue_names(self) -> list[str]:
        return self.agenda.issue_names

    def to_numeric(self) -> Scenario:
        """
        Forces all issues in the domain to become numeric

        Remarks:
            - maps the agenda and ufuns to work correctly together
        """
        raise NotImplementedError()

    def to_single_issue(self, numeric=False, stringify=True) -> Scenario:
        """
        Forces the domain to have a single issue with all possible outcomes

        Args:
            numeric: If given, the output issue will be a `ContiguousIssue` otherwise it will be a `DiscreteCategoricalIssue`
            stringify:  If given, the output issue will have string values. Checked only if `numeric` is `False`

        Remarks:
            - maps the agenda and ufuns to work correctly together
            - Only works if the outcome space is finite
        """
        if hasattr(self.agenda, "issues") and len(self.agenda.issues) == 1:
            return self
        outcomes = list(self.agenda.enumerate_or_sample())
        sos = self.agenda.to_single_issue(numeric, stringify)
        ufuns = []
        souts = list(sos.issues[0].all)
        for u in self.ufuns:
            if isinstance(u, DiscountedUtilityFunction):
                usave = u
                v = u.ufun
                while isinstance(v, DiscountedUtilityFunction):
                    u, v = v, v.ufun
                u.ufun = LinearAdditiveUtilityFunction(
                    values=(TableFun(dict(zip(souts, [v(_) for _ in outcomes]))),),
                    bias=0.0,
                    reserved_value=v.reserved_value,
                    name=v.name,
                    outcome_space=sos,
                )
                ufuns.append(usave)
                continue
            ufuns.append(
                LinearAdditiveUtilityFunction(
                    values=(TableFun(dict(zip(souts, [u(_) for _ in outcomes]))),),
                    bias=0.0,
                    reserved_value=u.reserved_value,
                    name=u.name,
                    outcome_space=sos,
                )
            )
        self.ufuns = tuple(ufuns)
        self.agenda = sos
        return self

    def make_session(
        self,
        negotiators: Callable[[], Negotiator] | list[Negotiator] | None = None,
        n_steps: int | None = None,
        time_limit: float | None = None,
        roles: list[str] | None = None,
        **kwargs,
    ):
        """
        Generates a ready to run mechanism session for this domain.
        """
        if not self.mechanism_type:
            raise ValueError(
                "Cannot create the domain because it has no `mechanism_type`"
            )

        args = self.mechanism_params
        args.update(kwargs)
        m = self.mechanism_type(
            outcome_space=self.agenda, n_steps=n_steps, time_limit=time_limit, **args
        )
        if not negotiators:
            return m
        negs: list[Negotiator]
        if not isinstance(negotiators, Iterable):
            negs = [negotiators() for _ in range(self.n_negotiators)]
        else:
            negs = list(negotiators)
        if len(self.ufuns) != len(negs) or len(negs) < 1:
            raise ValueError(
                f"Invalid ufuns ({self.ufuns}) or negotiators ({negotiators})"
            )
        if not roles:
            roles = ["negotiator"] * len(negs)
        for n, r, u in zip(negs, roles, self.ufuns):
            m.add(n, preferences=u, role=r)
        return m

    def scale_min(
        self,
        to: float = 1.0,
    ) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            ufun: The utility function to normalize
            outcomes: A collection of outcomes to normalize for
            rng: range to normalize to. Default is [0, 1]
            levels: Number of levels to use for discretizing continuous issues (if any)
            max_cardinality: Maximum allowed number of outcomes resulting after all discretization is done
        """
        self.ufuns = tuple(_.scale_min(to) for _ in self.ufuns)  # type: ignore The type is correct
        return self

    def scale_max(
        self,
        to: float = 1.0,
    ) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            ufun: The utility function to normalize
            outcomes: A collection of outcomes to normalize for
            rng: range to normalize to. Default is [0, 1]
            levels: Number of levels to use for discretizing continuous issues (if any)
            max_cardinality: Maximum allowed number of outcomes resulting after all discretization is done
        """
        self.ufuns = tuple(_.scale_max(to) for _ in self.ufuns)  # type: ignore The type is correct
        return self

    def normalize(
        self,
        to: tuple[float, float] = (0.0, 1.0),
    ) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            rng: range to normalize to. Default is [0, 1]
        """
        self.ufuns = tuple(_.normalize(to) for _ in self.ufuns)  # type: ignore The type is correct
        return self

    def discretize(self, levels: int = 10):
        """Discretize all issues"""
        self.agenda = self.agenda.to_discrete(levels)
        return self

    def remove_discounting(self):
        """Removes discounting from all ufuns"""
        ufuns = []
        for u in self.ufuns:
            while isinstance(u, DiscountedUtilityFunction):
                u = u.ufun
            ufuns.append(u)
        self.ufuns = tuple(ufuns)
        return self

    def remove_reserved_values(self, r: float = float("-inf")):
        """Removes reserved values from all ufuns replaacing it with `r`"""
        for u in self.ufuns:
            u.reserved_value = r
        return self

    def calc_stats(
        self, max_cardinality: int = STATS_MAX_CARDINALITY
    ) -> dict[str, Any]:
        """
        Calculates and returns several stats corresponding to the domain

        Args:
            max_cardinality (int): The maximum number of outcomes considered when calculating the stats.

        Returns:
            A dictionary with the compiled stats
        """
        agenda, ufuns = self.agenda, self.ufuns
        outcomes = tuple(agenda.enumerate_or_sample(max_cardinality=max_cardinality))
        frontier_utils, frontier_indices = pareto_frontier(
            ufuns, outcomes=outcomes, sort_by_welfare=True
        )
        frontier_outcomes = tuple(
            outcomes[_] for _ in frontier_indices if _ is not None
        )
        nash_utils, nash_indx = nash_point(ufuns, frontier_utils, outcome_space=agenda)
        nash_outcome = outcomes[nash_indx] if nash_indx else None
        minmax = [u.minmax() for u in ufuns]
        nu, no, ol, cl, wl, fu, fo = (
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
        )
        opposition = opposition_level(
            ufuns,
            max_utils=tuple(_[1] for _ in minmax),
            outcomes=outcomes,
            max_tests=max_cardinality,
        )
        for i, u1 in enumerate(ufuns):
            if not u1:
                continue
            for j, u2 in enumerate(ufuns[i + 1 :]):
                if not u2:
                    continue
                us = (u1, u2)
                fu_, findx = pareto_frontier(
                    us, outcomes=outcomes, sort_by_welfare=True
                )
                fu[(u1.name, u2.name)], fo[(u1.name, u2.name)] = fu_, [
                    outcomes[_] for _ in findx
                ]
                nu[(u1.name, u2.name)], nindx = nash_point(
                    (u1, u2), fu_, outcomes=outcomes
                )
                no[(u1.name, u2.name)] = outcomes[nindx] if nindx else None
                ol[(u1.name, u2.name)] = opposition_level(
                    (u1, u2),
                    outcomes=outcomes,
                    max_utils=(minmax[i][1], minmax[j][1]),
                    max_tests=max_cardinality,
                )
                cl[(u1.name, u2.name)] = conflict_level(
                    u1, u2, outcomes=outcomes, max_tests=max_cardinality
                )
                wl[(u1.name, u2.name)] = winwin_level(
                    u1, u2, outcomes=outcomes, max_tests=max_cardinality
                )

        return dict(
            frontier_utils=frontier_utils,
            frontier_outcomes=frontier_outcomes,
            nash_utils=nash_utils,
            nash_outcome=nash_outcome,
            opposition_level=opposition,
            bilateral_nash_utils=nu,
            bilateral_nash_outcome=no,
            bilateral_conflict_level=cl,
            bilateral_opposition_level=ol,
            bilateral_winwin_levl=wl,
            bilateral_frontier_utils=fu,
            bilateral_frontier_outcomes=fo,
        )

    def serialize(self) -> dict[str, Any]:
        """
        Converts the current scenario into a serializable dict.

        Remarks:
            Rturns a dictionary with the following keys:
                - domain: The agenda/outcome-space
                - ufuns: A list of utility functions
        """

        def get_name(x, default):
            if not x:
                return str(default)
            return x.split("/")[-1].replace(".xml", "")

        def adjust(
            d,
            default_name,
            remove_dunder=False,
            adjust_name=True,
            ignored=("id", "n_values", "outcome_space"),
            rename={PYTHON_CLASS_IDENTIFIER: "type"},
        ):
            if isinstance(d, list) or isinstance(d, tuple):
                return [
                    adjust(_, default_name, remove_dunder, adjust_name, ignored)
                    for _ in d
                ]
            if not isinstance(d, dict):
                return d
            if adjust_name and "name" in d:
                d["name"] = get_name(d["name"], default_name)
            if d.get(PYTHON_CLASS_IDENTIFIER, "").startswith("negmas."):
                d[PYTHON_CLASS_IDENTIFIER] = d[PYTHON_CLASS_IDENTIFIER].split(".")[-1]
            for old, new in rename.items():
                if old in d.keys():
                    d[new] = d[old]
                    del d[old]
            for i in ignored:
                if i in d.keys():
                    del d[i]
            for k, v in d.items():
                d[k] = adjust(
                    v,
                    default_name,
                    remove_dunder=remove_dunder,
                    adjust_name=False,
                    ignored=ignored,
                )
            if not remove_dunder:
                return d
            d = {k: v for k, v in d.items() if not k.startswith("__")}
            return d

        domain = adjust(
            serialize(self.agenda, shorten_type_field=True, add_type_field=True),
            "domain",
        )
        ufuns = [
            adjust(serialize(u, shorten_type_field=True, add_type_field=True), i)
            for i, u in enumerate(self.ufuns)
        ]
        return dict(domain=domain, ufuns=ufuns)

    def to_yaml(self, folder: Path | str) -> None:
        """
        Saves the scenario as yaml
        Args:
            folder: The destiation path
        """
        self.dumpas(folder, "yml")

    def to_json(self, folder: Path | str) -> None:
        """
        Saves the scenario as json
        Args:
            folder: The destiation path
        """
        self.dumpas(folder, "json")

    def dumpas(self, folder: Path | str, type="yml") -> None:
        """
        Dumps the scenrio in the given file format.
        """
        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)
        serialized = self.serialize()
        dump(serialized["domain"], folder / f"{serialized['domain']['name']}.{type}")
        for u in serialized["ufuns"]:
            dump(u, folder / f"{u['name']}.{type}")

    @staticmethod
    def from_genius_folder(
        path: PathLike | str,
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_genius_domain_from_folder(
            folder_name=str(path),
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            safe_parsing=safe_parsing,
        )

    @staticmethod
    def from_genius_files(
        domain: PathLike,
        ufuns: Iterable[PathLike],
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_genius_domain(
            domain,
            [_ for _ in ufuns],
            safe_parsing=safe_parsing,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            normalize_utilities=False,
            normalize_max_only=False,
        )


def get_domain_issues(
    domain_file_name: PathLike | str,
    n_discretization: int | None = None,
    safe_parsing=False,
) -> Sequence[Issue] | None:
    """
    Returns the issues of a given XML domain (Genius Format)

    Args:
        domain_file_name: Name of the file
        n_discretization: Number of discrete levels per continuous variable.
        max_cardinality: Maximum number of outcomes in the outcome space after discretization. Used only if `n_discretization` is given.
        safe_parsing: Apply more checks while parsing

    Returns:
        List of issues

    """
    issues = None
    if domain_file_name is not None:
        issues, _ = issues_from_genius(
            domain_file_name,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
    return issues


def load_genius_domain(
    domain_file_name: PathLike,
    utility_file_names: Iterable[PathLike] | None = None,
    ignore_discount=False,
    ignore_reserved=False,
    safe_parsing=True,
    **kwargs,
) -> Scenario:
    """
    Loads a genius domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name: XML file containing Genius-formatted domain spec
        utility_file_names: XML files containing Genius-fromatted ufun spec
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing
        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor

    Returns:
        A `Domain` ready to run

    """

    issues = None
    if domain_file_name is not None:
        issues, _ = issues_from_genius(
            domain_file_name,
            safe_parsing=safe_parsing,
        )

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    for ufname in utility_file_names:
        utility, discount_factor = UtilityFunction.from_genius(
            file_name=ufname,
            issues=issues,
            safe_parsing=safe_parsing,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            name=str(ufname),
        )
        agent_info.append(
            {
                "ufun": utility,
                "ufun_name": ufname,
                "reserved_value_func": utility.reserved_value
                if utility is not None
                else float("-inf"),
                "discount_factor": discount_factor,
            }
        )
    if domain_file_name is not None:
        kwargs["avoid_ultimatum"] = False
        kwargs["dynamic_entry"] = False
        kwargs["max_n_agents"] = None
        if not ignore_discount:
            for info in agent_info:
                info["ufun"] = (
                    info["ufun"]
                    if info["discount_factor"] is None or info["discount_factor"] == 1.0
                    else make_discounted_ufun(
                        ufun=info["ufun"],
                        discount_per_round=info["discount_factor"],
                        power_per_round=1.0,
                    )
                )
    if issues is None:
        raise ValueError(f"Could not load domain {domain_file_name}")

    return Scenario(
        agenda=make_os(issues, name=str(domain_file_name)),
        ufuns=[_["ufun"] for _ in agent_info],  # type: ignore
        mechanism_type=SAOMechanism,
        mechanism_params=kwargs,
    )


def load_genius_domain_from_folder(
    folder_name: str | PathLike,
    ignore_reserved=False,
    ignore_discount=False,
    safe_parsing=False,
    **kwargs,
) -> Scenario:
    """
    Loads a genius domain from a folder. See ``load_genius_domain`` for more details.

    Args:
        folder_name: A folder containing one XML domain file and one or more ufun files in Genius format
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing
        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor

    Returns:
        A domain ready for `make_session`

    Examples:

        >>> import pkg_resources
        >>> from negmas import load_genius_domain_from_folder
        >>> from negmas import AspirationNegotiator

        Try loading and running a domain with predetermined agents:
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop'))
        >>> mechanism = domain.make_session(AspirationNegotiator, n_steps=100)
        >>> state = mechanism.run()
        >>> state.agreement is not None
        True


        Try loading a domain and check the resulting ufuns
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop'))

        >>> domain.n_issues, domain.n_negotiators
        (3, 2)

        >>> [type(_) for _ in domain.ufuns]
        [<class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>, <class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>]

        Try loading a domain forcing a single issue space
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ... ).to_single_issue()
        >>> domain.n_issues, domain.n_negotiators
        (1, 2)
        >>> [type(_) for _ in domain.ufuns]
        [<class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>, <class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>]


        Try loading a domain with nonlinear ufuns:
        >>> folder_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/10issues')
        >>> domain = load_genius_domain_from_folder(folder_name)
        >>> print(domain.n_issues)
        10
        >>> print(domain.n_negotiators)
        2
        >>> print([type(u) for u in domain.ufuns])
        [<class 'negmas.preferences.crisp.nonlinear.HyperRectangleUtilityFunction'>, <class 'negmas.preferences.crisp.nonlinear.HyperRectangleUtilityFunction'>]
        >>> u = domain.ufuns[0]
        >>> print(u.outcome_ranges[0])
        {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}

        >>> print(u.mappings[0])
        97.0
        >>> print(u([0.0] * domain.n_issues))
        0
        >>> print(u([0.5] * domain.n_issues))
        186.0
    """
    folder_name = str(folder_name)
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = Path(full_name)
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    if domain_file_name is None:
        raise ValueError("Cannot find a domain file")
    return load_genius_domain(
        domain_file_name=domain_file_name,
        utility_file_names=utility_file_names,
        safe_parsing=safe_parsing,
        ignore_reserved=ignore_reserved,
        ignore_discount=ignore_discount,
        **kwargs,
    )


def find_domain_and_utility_files(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from negmas.sao.components.concession import KindConcessionRecommender

from ..components import TFTAcceptanceStrategy, TFTOfferingStrategy, ZeroSumModel
from .modular import MAPNegotiator

if TYPE_CHECKING:
    pass


__all__ = [
    "NaiveTitForTatNegotiator",
    "SimpleTitForTatNegotiator",
]


class NaiveTitForTatNegotiator(MAPNegotiator):
    """
    Implements a naive tit-for-tat strategy that does not depend on the availability of an opponent model.

    Args:
        name: Negotiator name
        preferences: negotiator preferences
        ufun: negotiator ufun (overrides preferences)
        parent: A controller
        kindness: How 'kind' is the agent. A value of zero is standard tit-for-tat. Positive values makes the negotiator
                  concede faster and negative values slower.
        stochastic: If `True`, the offers will be randomized above the level determined by the current concession
                        which in turn reflects the opponent's concession.
        punish: If `True` the agent punish a partner who does not seem to conede by requiring higher utilities
        initial_concession: How much should the agent concede in the beginning in terms of utility. Should be a number
                            or the special string value 'min' for minimum concession

    Remarks:
        - This negotiator does not keep an opponent model. It thinks only in terms of changes in its own utility.
          If the opponent's last offer was better for the negotiator compared with the one before it, it considers
          that the opponent has conceded by the difference. This means that it implicitly assumes a zero-sum
          situation.
    """

    def __init__(
        self,
        *args,
        kindness=0.0,
        punish=False,
        initial_concession: float | Literal["min"] = "min",
        rank_only: bool = False,
        stochastic: bool = False,
        **kwargs,
    ):
        partner_model = ZeroSumModel(rank_only=rank_only, above_reserve=False)
        if isinstance(initial_concession, str):
            initial_concession = 0
        recommender = KindConcessionRecommender(
            initial_concession=initial_concession, kindness=kindness, punish=punish
        )
        acceptance = TFTAcceptanceStrategy(
            recommender=recommender, partner_ufun=partner_model
        )
        offering = TFTOfferingStrategy(
            recommender=recommender, partner_ufun=partner_model, stochastic=stochastic
        )
        super().__init__(
            *args,
            models=[partner_model, recommender],
            model_names=["partner-model", "concession-recommender"],
            acceptance=acceptance,
            offering=offering,
            **kwargs,
        )


SimpleTitForTatNegotiator = NaiveTitForTatNegotiator
"""A simple tit-for-tat negotiator based on the MAP architecture"""



from __future__ import annotations

from ..components import AcceptAbove, AcceptBest, AcceptTop, OfferBest, OfferTop
from .modular import MAPNegotiator

__all__ = [
    "ToughNegotiator",
    "TopFractionNegotiator",
]


class ToughNegotiator(MAPNegotiator):
    """
    Accepts and proposes only the top offer (i.e. the one with highest utility).

    Args:
         name: Negotiator name
         parent: Parent controller if any
         can_propose: If `False` the negotiator will never propose but can only accept
         preferences: The preferences of the negotiator
         ufun: The ufun of the negotiator (overrides preferences)
         owner: The `Agent` that owns the negotiator.

    Remarks:
        - If there are multiple outcome with the same maximum utility, only one of them will be used.

    """

    def __init__(
        self,
        can_propose=True,
        **kwargs,
    ):
        acceptance = AcceptBest()
        offering = None if not can_propose else OfferBest()
        super().__init__(acceptance=acceptance, offering=offering, **kwargs)


class TopFractionNegotiator(MAPNegotiator):
    """
    Offers and accepts only one of the top outcomes for the negotiator.

    Args:
        name: Negotiator name
        parent: Parent controller if any
        can_propose: If `False` the negotiator will never propose but can only accept
        preferences: The preferences of the negotiator
        ufun: The ufun of the negotiator (overrides preferences)
        min_utility: The minimum utility to offer or accept
        top_fraction: The fraction of the outcomes (ordered decreasingly by utility) to offer or accept
        best_first: Guarantee offering will non-increasing in terms of utility value
        probabilistic_offering: Offer randomly from the outcomes selected based on `top_fraction` and `min_utility`
        owner: The `Agent` that owns the negotiator.
    """

    def __init__(
        self,
        min_utility=0.95,
        top_fraction=0.05,
        best_first=True,
        can_propose=True,
        **kwargs,
    ):
        acceptance = AcceptAbove(min_utility) and AcceptTop(top_fraction)
        offering = None if not can_propose else OfferTop(top_fraction)
        self._best_first = best_first
        self.__offered = False
        super().__init__(acceptance=acceptance, offering=offering, **kwargs)

    def propose(self, state):
        if not self.ufun:
            return None
        if not self.__offered:
            _, best = self.ufun.extreme_outcomes()
            self.__offered = True
            return best
        return super().propose(state)



"""
Functions for handling outcome spaces
"""
from __future__ import annotations

import math
import numbers
from typing import TYPE_CHECKING, Any, Iterable, Sequence, overload

import numpy as np

from negmas.generics import ienumerate, iget, ikeys
from negmas.helpers.numeric import isint, isreal
from negmas.outcomes.cardinal_issue import CardinalIssue
from negmas.outcomes.range_issue import RangeIssue

from .base_issue import Issue

if TYPE_CHECKING:
    from .common import Outcome, OutcomeRange
    from .outcome_space import DistanceFun, OutcomeSpace

__all__ = [
    "dict2outcome",
    "outcome2dict",
    "outcome_in_range",
    "outcome_is_complete",
    "outcome_types_are_ok",
    "outcome_is_valid",
    "generalized_minkowski_distance",
    "min_dist",
]


def _is_single(x):
    """Checks whether a value is a single value which is defined as either a string or not an Iterable."""

    return isinstance(x, str) or isinstance(x, numbers.Number)


@overload
def outcome2dict(outcome: None, issues: Sequence[str | Issue]) -> None:
    ...


@overload
def outcome2dict(outcome: Outcome, issues: Sequence[str | Issue]) -> dict[str, Any]:
    ...


def outcome2dict(
    outcome: Outcome | None, issues: Sequence[str | Issue]
) -> dict[str, Any] | None:
    """
    Converts the outcome to a dict no matter what was its type.


    Args:

        outcome: The outcome to be converted (as a tuple)
        issues: The issues/issue names used as dictionary keys in the output


    Remarks:
        - If called with a dict that is already converted, it will just return it.
        - None is converted to None

    Examples:

        >>> from negmas import make_issue
        >>> issues = [make_issue(10, "price"), make_issue(5, "quantity")]
        >>> outcome2dict((3, 4), issues=issues)
        {'price': 3, 'quantity': 4}

        You can also use issue names without creating Issue objects
        >>> issues = ["price", "quantity"]
        >>> outcome2dict((3, 4), issues=issues)
        {'price': 3, 'quantity': 4}

        Trying to convert an already converted outcome does nothing
        >>> issues = ["price", "quantity"]
        >>> outcome2dict(outcome2dict((3, 4), issues=issues), issues=issues)
        {'price': 3, 'quantity': 4}
    """

    if outcome is None:
        return None

    # if len(outcome) != len(issues):
    #     raise ValueError(
    #         f"Cannot convert {len(outcome)} valued outcome to a dict with {len(issues)} issues"
    #     )

    if isinstance(outcome, np.ndarray):
        outcome = tuple(outcome.tolist())

    if isinstance(outcome, dict):
        names = {_.name if isinstance(_, Issue) else _ for _ in issues}
        for k in outcome.keys():
            if k not in names:
                raise ValueError(
                    f"{k} not in the issue names ({names}). An invalid dict is given!!"
                )
        return outcome

    return dict(zip([_ if isinstance(_, str) else _.name for _ in issues], outcome))


def dict2outcome(d: dict[str, Any] | None, issues: list[str | Issue]) -> Outcome | None:
    """
    Converts the outcome to a tuple no matter what was its type

    Args:
        d: the dictionary to be converted
        issues: A list of issues or issue names (as strings) to order the tuple

    Remarks:
        - If called with a tuple outcome, it will issue a warning

    """

    if d is None:
        return None

    if isinstance(d, tuple):
        return d

    return tuple(d[_ if isinstance(_, str) else _.name] for _ in issues)


def generalized_minkowski_distance(
    a: Outcome,
    b: Outcome,
    outcome_space: OutcomeSpace | None,
    *,
    weights: Sequence[float] | None = None,
    dist_power: float = 2,
) -> float:
    r"""
    Calculates the difference between two outcomes given an outcome-space (optionally with issue weights). This is defined as the distance.

    Args:

        outcome_space: The outcome space used for comparison (If None an apporximate implementation is provided)
        a: first outcome
        b: second outcome
        weights: Issue weights
        dist_power: The exponent used when calculating the distance

    Remarks:

        - Implements the following distance measure:

          .. math::

             d(a, b) = \left( \sum_{i=1}^{N} w_i {\left| a_i - b_i \right|}^p \right)^{frac{1}{p}}

          where $a, b$ are the outocmes, $x_i$ is value for issue $i$ of outcoem $x$, $w_i$ is the weight of issue $i$ and $p$ is the `dist_power` passsed. Categorical issue differences is defined as $1$ if the values are not
          equal and $0$ otherwise.
        - Becomes the Euclidean distance if all issues are numeric and no weights are given
        - You can control the power:

            - Setting it to 1 is the city-block distance
            - Setting it to 0 is the maximum issue difference
    """
    from negmas.outcomes import CartesianOutcomeSpace

    if not weights:
        weights = [1] * len(a)
    if dist_power <= 0 or dist_power == float("inf"):
        if not isinstance(outcome_space, CartesianOutcomeSpace):
            return max(
                (w * abs(x - y))
                if (isint(x) or isreal(x)) and (isint(y) or isreal(y))
                else (w * int(x == y))
                for w, x, y in zip(weights, a, b)
            )
        d = float("-inf")
        for issue, w, x, y in zip(outcome_space.issues, weights, a, b):
            if isinstance(issue, CardinalIssue):
                c = w * abs(x - y)
            else:
                c = w * int(x == y)
            if c > d:
                d = c
        return d
    if not isinstance(outcome_space, CartesianOutcomeSpace):
        return math.pow(
            sum(
                (w * math.pow(abs(x - y), dist_power))
                if (isint(x) or isreal(x)) and (isint(y) or isreal(y))
                else (w * int(x == y))
                for w, x, y in zip(weights, a, b)
            ),
            1.0 / dist_power,
        )
    d = 0.0
    for issue, w, x, y in zip(outcome_space.issues, weights, a, b):
        if isinstance(issue, CardinalIssue):
            d += w * math.pow(abs(x - y), dist_power)
            continue
        d += w * int(x == y)
    return math.pow(d, 1.0 / dist_power)


def min_dist(
    test_outcome: Outcome,
    outcomes: Sequence[Outcome],
    outcome_space: OutcomeSpace | None,
    distance_fun: DistanceFun = generalized_minkowski_distance,
    **kwargs,
) -> float:
    """
    Minimum distance between an outcome and a set of outcomes in an outcome-spaceself.

    Args:
        test_outcome: The outcome tested
        outcomes: A sequence of outcomes to compare to
        outcome_space: The outcomespace used for comparison
        distance_fun: The distance function
        kwargs: Paramters to pass to the distance function


    See Also:

        `generalized_euclidean_distance`
    """
    if not outcomes:
        return 1.0
    return min(distance_fun(test_outcome, _, outcome_space, **kwargs) for _ in outcomes)


def outcome_is_valid(outcome: Outcome, issues: Iterable[Issue]) -> bool:
    """
    Test validity of an outcome given a set of issues.

    Examples:

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue((0.5, 2.0), 'price'), make_issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , make_issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: (0, 19)
        >>> print([outcome_is_valid({'price':3.0}, issues), outcome_is_valid({'date': '2018.10.4'}, issues)\
            , outcome_is_valid({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_valid(valid_incomplete, issues))
        True
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True

    Args:
        outcome: outcome tested which can contain values for a partial set of issue values
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure
    """
    outcome_dict = outcome2dict(outcome, [_.name for _ in issues])

    for issue in issues:
        for key in outcome_dict.keys():
            if str(issue.name) == str(key):
                break

        else:
            continue

        value = iget(outcome_dict, key)

        if isinstance(issue, RangeIssue) and (
            isinstance(value, str) or not issue.min_value <= value <= issue.max_value
        ):
            return False

        if isinstance(issue._values, list) and value not in issue._values:
            return False

    return True


def outcome_types_are_ok(outcome: Outcome, issues: Iterable[Issue]) -> bool:
    """
    Checks that the types of all issue values in the outcome are correct
    """
    if not issues or not outcome:
        return True
    for v, i in zip(outcome, issues):
        if i.value_type is None:
            continue
        if not isinstance(v, i.value_type):
            return False
    return True


def cast_value_types(outcome: Outcome, issues: Iterable[Issue]) -> Outcome:
    """
    Casts the types of values in the outcomes to the value-type of each issue (if given)
    """
    if not issues or not outcome:
        return outcome
    new_outcome = list(outcome)
    for indx, (v, i) in enumerate(zip(outcome, issues)):
        if i.value_type is None:
            continue
        new_outcome[indx] = i.value_type(v)  # type: ignore I know that value_type is callable
    return tuple(new_outcome)


def outcome_is_complete(outcome: Outcome, issues: Sequence[Issue]) -> bool:
    """
    Tests that the outcome is valid and complete.

    Examples:

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue((0.5, 2.0), 'price'), make_issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')\
                , make_issue(20, 'count')]
        >>> for _ in issues: print(_)
        price: (0.5, 2.0)
        date: ['2018.10.1', '2018.10.2', '2018.10.3']
        count: (0, 19)
        >>> print([outcome_is_complete({'price':3.0}, issues), outcome_is_complete({'date': '2018.10.4'}, issues)\
            , outcome_is_complete({'count': 21}, issues)])
        [False, False, False]
        >>> valid_incomplete = {'price': 1.9}
        >>> print(outcome_is_complete(valid_incomplete, issues))
        False
        >>> valid_incomplete.update({'date': '2018.10.2', 'count': 5})
        >>> print(outcome_is_complete(valid_incomplete, issues))
        True
        >>> invalid = {'price': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False
        >>> invalid = {'unknown': 2000, 'date': '2018.10.2', 'count': 5}
        >>> print(outcome_is_complete(invalid, issues))
        False

     Args:
        outcome: outcome tested which much contain valid values all issues if it is to be considered complete.
        issues: issues

    Returns:
        Union[bool, Tuple[bool, str]]: If return_problem is True then a second return value contains a string with
                                      reason of failure

    """
    try:
        outcome2dict(outcome, issues)
    except ValueError:
        return False

    if len(outcome) != len(issues):
        return False

    valid = outcome_is_valid(outcome, issues)

    if not valid:
        return False

    outcome_keys = [str(k) for k in ikeys(outcome)]

    for issue in issues:
        if str(issue.name) not in outcome_keys:
            return False

    return True


def outcome_in_range(
    outcome: Outcome,
    outcome_range: OutcomeRange,
    *,
    strict=False,
    fail_incomplete=False,
) -> bool:
    """
    Tests that the outcome is contained within the given range of outcomes.

    An outcome range defines a value or a range of values for each issue.

    Args:

        outcome: "Outcome" being tested
        outcome_range: "Outcome" range being tested against
        strict: Whether to enforce that all issues in the outcome must be mentioned in the outcome_range
        fail_incomplete: If True then outcomes that do not sepcify a value for all keys in the outcome_range
        will be considered not falling within it. If False then these outcomes will be considered falling
        within the range given that the values for the issues mentioned in the outcome satisfy the range
        constraints.

    Examples:

        >>> outcome_range = {'price': (0.0, 2.0), 'distance': [0.3, 0.4], 'type': ['a', 'b'], 'area': 3}
        >>> outcome_range_2 = {'price': [(0.0, 1.0), (1.5, 2.0)], 'area': [(3, 4), (7, 9)]}
        >>> outcome_in_range({'price':3.0}, outcome_range)
        False
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range, strict=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range, fail_incomplete=True)
        False
        >>> outcome_in_range({'area': 3}, outcome_range)
        True
        >>> outcome_in_range({'type': 'c'}, outcome_range)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range)
        True
        >>> outcome_in_range({'date': '2018.10.4'}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3.1}, outcome_range_2)
        True
        >>> outcome_in_range({'area': 3}, outcome_range_2)
        False
        >>> outcome_in_range({'area': 5}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4}, outcome_range_2, fail_incomplete=True)
        False
        >>> outcome_in_range({'price': 1.2}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 0.4, 'area': 3.9}, outcome_range_2)
        True
        >>> outcome_in_range({'price': 0.4, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 10}, outcome_range_2)
        False
        >>> outcome_in_range({'price': 1.2, 'area': 4}, outcome_range_2)
        False
        >>> outcome_in_range({'type': 'a'}, outcome_range_2)
        True
        >>> outcome_in_range({'type': 'a'}, outcome_range_2, strict=True)
        False
        >>> outcome_range = {'price': 10}
        >>> outcome_in_range({'price': 10}, outcome_range)
        True
        >>> outcome_in_range({'price': 11}, outcome_range)
        False

    Returns:

        bool: Success or failure

    Remarks:
        Outcome ranges specify regions in an outcome space. They can have any of the following conditions:

        - A key/issue not mentioned in the outcome range does not add any constraints meaning that **All**
          values are acceptable except if strict == True. If strict == True then *NO* value will be accepted for issues
          not in the outcome_range.
        - A key/issue with the value None in the outcome range means **All** values on this issue are acceptable.
          This is the same as having this key/issue removed from the outcome space
        - A key/issue withe the value [] (empty list) accepts *NO* outcomes
        - A key/issue with  a single value means that it is the only one acceptable
        - A key/issue with a single 2-items tuple (min, max) means that any value within that range is acceptable.
        - A key/issue with a list of values means an output is acceptable if it falls within the condition specified
          by any of the values in the list (list == union). Each such value can be a single value, a 2-items
          tuple or another list. Notice that lists of lists can always be combined into a single list of values

    """

    if (
        fail_incomplete
        and len(set(ikeys(outcome_range)).difference(ikeys(outcome))) > 0
    ):
        return False

    for key, value in ienumerate(outcome):
        if key not in ikeys(outcome_range):
            if strict:
                return False

            continue

        values = iget(outcome_range, key, None)

        if values is None:
            return False

        if _is_single(values) and value != values:
            return False

        if isinstance(values, tuple) and not values[0] < value < values[1]:
            return False

        if isinstance(values, list):
            for constraint in values:
                if _is_single(constraint):
                    if value == constraint:
                        break

                elif isinstance(constraint, list):
                    if value in constraint:
                        break

                elif isinstance(constraint, tuple):
                    if constraint[0] < value < constraint[1]:
                        break

            else:
                return False

            continue

    return True



from __future__ import annotations

import random
from functools import reduce
from operator import mul
from typing import Callable, Iterable, Sequence, Union

from attr import define, field

from negmas.helpers import unique_name
from negmas.helpers.types import get_full_type_name
from negmas.outcomes.outcome_ops import (
    cast_value_types,
    outcome_is_valid,
    outcome_types_are_ok,
)
from negmas.protocols import XmlSerializable
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize
from negmas.warnings import NegmasSpeedWarning, warn

from .base_issue import DiscreteIssue, Issue
from .categorical_issue import CategoricalIssue
from .common import Outcome
from .contiguous_issue import ContiguousIssue
from .issue_ops import (
    enumerate_discrete_issues,
    issues_from_outcomes,
    issues_from_xml_str,
    issues_to_xml_str,
    sample_issues,
)
from .protocols import DiscreteOutcomeSpace, OutcomeSpace
from .range_issue import RangeIssue

__all__ = [
    "CartesianOutcomeSpace",
    "DiscreteCartesianOutcomeSpace",
    "make_os",
    "DistanceFun",
]

NLEVELS = 5


DistanceFun = Callable[[Outcome, Outcome, Union[OutcomeSpace, None]], float]
"""A callable that can calculate the distance between two outcomes in an outcome-space"""


def make_os(
    issues: Sequence[Issue] | None = None,
    outcomes: Sequence[Outcome] | None = None,
    name: str | None = None,
) -> CartesianOutcomeSpace:
    """
    A factory to create outcome-spaces from lists of `Issue` s or `Outcome` s.

    Remarks:

        - must pass one and exactly one of `issues` and `outcomes`
    """
    if issues and outcomes:
        raise ValueError(
            f"Cannot make an outcome space passing both issues and outcomes"
        )
    if not issues and not outcomes:
        raise ValueError(
            f"Cannot make an outcome space without passing issues or outcomes"
        )
    if not issues and outcomes:
        issues_ = issues_from_outcomes(outcomes)
    else:
        issues_ = issues
    if issues_ is None:
        raise ValueError(
            f"Cannot make an outcome space without passing issues or outcomes"
        )

    issues_ = tuple(issues_)
    if all(_.is_discrete() for _ in issues_):
        return DiscreteCartesianOutcomeSpace(issues_, name=name if name else "")
    return CartesianOutcomeSpace(issues_, name=name if name else "")


@define(frozen=True)
class CartesianOutcomeSpace(XmlSerializable):
    """
    An outcome-space that is generated by the cartesian product of a tuple of `Issue` s.
    """

    issues: tuple[Issue] = field(converter=tuple)
    name: str | None = field(eq=False, default=None)

    def __attrs_post_init__(self):
        if not self.name:
            object.__setattr__(self, "name", unique_name("os", add_time=False, sep=""))

    def contains_issue(self, x: Issue) -> bool:
        """Cheks that the given issue is in the tuple of issues constituting the outcome space (i.e. it is one of its dimensions)"""
        return x in self.issues

    def is_valid(self, outcome: Outcome) -> bool:
        return outcome_is_valid(outcome, self.issues)

    def is_discrete(self) -> bool:
        """Checks whether all issues are discrete"""
        return all(_.is_discrete() for _ in self.issues)

    def is_finite(self) -> bool:
        """Checks whether the space is finite"""
        return self.is_discrete()

    def contains_os(self, x: OutcomeSpace) -> bool:
        """Checks whether an outcome-space is contained in this outcome-space"""
        if isinstance(x, CartesianOutcomeSpace):
            return len(self.issues) == len(x.issues) and all(
                b in a for a, b in zip(self.issues, x.issues)
            )
        if self.is_finite() and not x.is_finite():
            return False
        if not self.is_finite() and not x.is_finite():
            raise NotImplementedError(
                "Cannot check an infinite outcome space that is not cartesian for inclusion in an infinite cartesian outcome space!!"
            )
        warn(
            f"Testing inclusion of a finite non-carteisan outcome space in a cartesian outcome space can be very slow (will do {x.cardinality} checks)",
            NegmasSpeedWarning,
        )
        return all(self.is_valid(_) for _ in x.enumerate())  # type: ignore If we are here, we know that x is finite

    def to_dict(self):
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            name=self.name,
            issues=serialize(self.issues),
        )

    @classmethod
    def from_dict(cls, d):
        return cls(**deserialize(d))  # type: ignore

    @property
    def issue_names(self) -> list[str]:
        """Returns an ordered list of issue names"""
        return [_.name for _ in self.issues]

    @property
    def cardinality(self) -> int | float:
        """The space cardinality = the number of outcomes"""
        return reduce(mul, [_.cardinality for _ in self.issues], 1)

    def is_compact(self) -> bool:
        """Checks whether all issues are complete ranges"""
        return all(isinstance(_, RangeIssue) for _ in self.issues)

    def is_all_continuous(self) -> bool:
        """Checks whether all issues are discrete"""
        return all(_.is_continuous() for _ in self.issues)

    def is_not_discrete(self) -> bool:
        """Checks whether all issues are discrete"""
        return any(_.is_continuous() for _ in self.issues)

    def is_numeric(self) -> bool:
        """Checks whether all issues are numeric"""
        return all(_.is_numeric() for _ in self.issues)

    def is_integer(self) -> bool:
        """Checks whether all issues are integer"""
        return all(_.is_integer() for _ in self.issues)

    def is_float(self) -> bool:
        """Checks whether all issues are real"""
        return all(_.is_float() for _ in self.issues)

    def to_discrete(
        self, levels: int | float = 10, max_cardinality: int | float = float("inf")
    ) -> DiscreteCartesianOutcomeSpace:
        """
        Discretizes the outcome space by sampling `levels` values for each continuous issue.

        The result of the discretization is stable in the sense that repeated calls will return the same output.
        """
        if max_cardinality != float("inf"):
            c = reduce(
                mul,
                [_.cardinality if _.is_discrete() else levels for _ in self.issues],
                1,
            )
            if c > max_cardinality:
                raise ValueError(
                    f"Cannot convert OutcomeSpace to a discrete OutcomeSpace with at most {max_cardinality} (at least {c} outcomes are required)"
                )
        issues = tuple(
            issue.to_discrete(
                levels if issue.is_continuous() else None,
                compact=False,
                grid=True,
                endpoints=True,
            )
            for issue in self.issues
        )
        return DiscreteCartesianOutcomeSpace(issues=issues, name=self.name)

    @classmethod
    def from_xml_str(
        cls, xml_str: str, safe_parsing=True, name=None
    ) -> CartesianOutcomeSpace:
        issues, _ = issues_from_xml_str(
            xml_str,
            safe_parsing=safe_parsing,
            n_discretization=None,
        )
        if not issues:
            raise ValueError(f"Failed to read an issue space from an xml string")
        issues = tuple(issues)
        if all(isinstance(_, DiscreteIssue) for _ in issues):
            return DiscreteCartesianOutcomeSpace(issues, name=name)
        return cls(issues, name=name)

    @staticmethod
    def from_outcomes(
        outcomes: list[Outcome],
        numeric_as_ranges: bool = False,
        issue_names: list[str] | None = None,
        name: str | None = None,
    ) -> DiscreteCartesianOutcomeSpace:
        return DiscreteCartesianOutcomeSpace(
            issues_from_outcomes(outcomes, numeric_as_ranges, issue_names), name=name
        )

    def to_xml_str(self) -> str:
        return issues_to_xml_str(self.issues)

    def are_types_ok(self, outcome: Outcome) -> bool:
        """Checks if the type of each value in the outcome is correct for the given issue"""
        return outcome_types_are_ok(outcome, self.issues)

    def ensure_correct_types(self, outcome: Outcome) -> Outcome:
        """Returns an outcome that is guaratneed to have correct types or raises an exception"""
        return cast_value_types(outcome, self.issues)

    def sample(
        self,
        n_outcomes: int,
        with_replacement: bool = True,
        fail_if_not_enough=True,
    ) -> Iterable[Outcome]:
        return sample_issues(
            self.issues, n_outcomes, with_replacement, fail_if_not_enough
        )

    def cardinality_if_discretized(
        self, levels: int, max_cardinality: int | float = float("inf")
    ) -> int:
        c = reduce(
            mul,
            [_.cardinality if _.is_discrete() else levels for _ in self.issues],
            1,
        )
        return min(c, max_cardinality)

    def to_largest_discrete(
        self, levels: int, max_cardinality: int | float = float("inf"), **kwargs
    ) -> DiscreteCartesianOutcomeSpace:
        for l in range(levels, 0, -1):
            if self.cardinality_if_discretized(levels) < max_cardinality:
                break
        else:
            raise ValueError(
                f"Cannot discretize with levels <= {levels} keeping the cardinality under {max_cardinality} Outocme space cardinality is {self.cardinality}\nOutcome space: {self}"
            )
        return self.to_discrete(l, max_cardinality, **kwargs)

    def enumerate_or_sample(
        self,
        levels: int | float = float("inf"),
        max_cardinality: int | float = float("inf"),
    ) -> Iterable[Outcome]:
        """Enumerates all outcomes if possible (i.e. discrete space) or returns `max_cardinality` different outcomes otherwise"""
        if (
            levels == float("inf")
            and max_cardinality == float("inf")
            and not self.is_discrete()
        ):
            raise ValueError(
                "Cannot enumerate-or-sample an outcome space with infinite outcomes without specifying `levels` and/or `max_cardinality`"
            )
        from negmas.outcomes.outcome_space import DiscreteCartesianOutcomeSpace

        if isinstance(self, DiscreteCartesianOutcomeSpace):
            return self.enumerate()  # type: ignore We know the outcome space is correct
        if max_cardinality == float("inf"):
            return self.to_discrete(
                levels=levels, max_cardinality=max_cardinality
            ).enumerate()
        return self.sample(
            int(max_cardinality), with_replacement=False, fail_if_not_enough=False
        )

    def to_single_issue(
        self,
        numeric=False,
        stringify=True,
        levels: int = NLEVELS,
        max_cardinality: int | float = float("inf"),
    ) -> DiscreteCartesianOutcomeSpace:
        """
        Creates a new outcome space that is a single-issue version of this one discretizing it as needed

        Args:
            numeric: If given, the output issue will be a `ContiguousIssue` otberwise it will be a `CategoricalIssue`
            stringify:  If given, the output issue will have string values. Checked only if `numeric` is `False`
            levels: Number of levels to discretize any continuous issue

        Remarks:
            - Will discretize inifinte outcome spaces
        """
        if isinstance(self, DiscreteCartesianOutcomeSpace) and len(self.issues) == 1:
            return self
        dos = self.to_discrete(levels, max_cardinality)
        return dos.to_single_issue(numeric, stringify)  # type: ignore

    def __contains__(self, item):
        if isinstance(item, OutcomeSpace):
            return self.contains_os(item)
        if isinstance(item, Issue):
            return self.contains_issue(item)
        if isinstance(item, Outcome):
            return self.is_valid(item)
        if not isinstance(item, Sequence):
            return False
        if not item:
            return True
        if isinstance(item[0], Issue):
            return len(self.issues) == len(item) and self.contains_os(
                make_os(issues=item)
            )
        if isinstance(item[0], Outcome):
            return len(self.issues) == len(item) and self.contains_os(
                make_os(outcomes=item)
            )
        return False


@define(frozen=True)
class DiscreteCartesianOutcomeSpace(CartesianOutcomeSpace):
    """
    A discrete outcome-space that is generated by the cartesian product of a tuple of `Issue` s (i.e. with finite number of outcomes).
    """

    def __attrs_post_init__(self):
        for issue in self.issues:
            if not issue.is_discrete():
                raise ValueError(
                    f"Issue is not discrete. Cannot be added to a DiscreteOutcomeSpace. You must discretize it first: {issue} "
                )

    @property
    def cardinality(self) -> int:
        return reduce(mul, [_.cardinality for _ in self.issues], 1)

    def cardinality_if_discretized(
        self, levels: int, max_cardinality: int | float = float("inf")
    ) -> int:
        return self.cardinality

    def enumerate(self) -> Iterable[Outcome]:
        return enumerate_discrete_issues(  #  type: ignore I know that all my issues are actually discrete
            self.issues  #  type: ignore I know that all my issues are actually discrete
        )

    def limit_cardinality(
        self,
        max_cardinality: int | float = float("inf"),
        levels: int | float = float("inf"),
    ) -> DiscreteCartesianOutcomeSpace:
        """
        Limits the cardinality of the outcome space to the given maximum (or the number of levels for each issue to `levels`)

        Args:
            max_cardinality: The maximum number of outcomes in the resulting space
            levels: The maximum number of levels for each issue/subissue
        """
        if self.cardinality <= max_cardinality or all(
            _.cardinality < levels for _ in self.issues
        ):
            return self
        new_levels = [_.cardinality for _ in self.issues]  # type: ignore will be corrected the next line
        new_levels = [int(_) if _ < levels else int(levels) for _ in new_levels]
        new_cardinality = reduce(mul, new_levels, 1)

        def _reduce_total_cardinality(new_levels, max_cardinality, new_cardinality):
            sort = reversed(sorted((_, i) for i, _ in enumerate(new_levels)))
            sorted_levels = [_[0] for _ in sort]
            indices = [_[1] for _ in sort]
            needed = new_cardinality - max_cardinality
            current = 0
            n = len(sorted_levels)
            while needed > 0 and current < n:
                nxt = n - 1
                v = sorted_levels[current]
                if v == 1:
                    continue
                for i in range(current + 1, n - 1):
                    if v == sorted_levels[i]:
                        continue
                    nxt = i
                    break
                diff = v - sorted_levels[nxt]
                if not diff:
                    diff = 1
                new_levels[indices[current]] -= 1
                max_cardinality = (max_cardinality // v) * (v - 1)
                sort = reversed(sorted((_, i) for i, _ in enumerate(new_levels)))
                sorted_levels = [_[0] for _ in sort]
                current = 0
                needed = new_cardinality - max_cardinality
            return new_levels

        if new_cardinality > max_cardinality:
            new_levels: list[int] = _reduce_total_cardinality(
                new_levels, max_cardinality, new_cardinality
            )
        issues: list[Issue] = []
        for j, i, issue in zip(
            new_levels, (_.cardinality for _ in self.issues), self.issues
        ):
            issues.append(issue if j >= i else issue.to_discrete(j, compact=True))
        return DiscreteCartesianOutcomeSpace(
            tuple(issues), name=f"{self.name}-{max_cardinality}"
        )

    def is_discrete(self) -> bool:
        """Checks whether there are no continua components of the space"""
        return True

    def to_discrete(self, *args, **kwargs) -> DiscreteOutcomeSpace:
        return self

    def to_single_issue(
        self, numeric=False, stringify=True
    ) -> DiscreteCartesianOutcomeSpace:
        """
        Creates a new outcome space that is a single-issue version of this one

        Args:
            numeric: If given, the output issue will be a `ContiguousIssue` otherwise it will be a `CategoricalIssue`
            stringify:  If given, the output issue will have string values. Checked only if `numeric` is `False`

        Remarks:
            - maps the agenda and ufuns to work correctly together
            - Only works if the outcome space is finite
        """
        outcomes = list(self.enumerate())
        values = (
            range(len(outcomes))
            if numeric
            else [f"v{_}" for _ in range(len(outcomes))]
            if stringify
            else outcomes
        )
        issue = (
            ContiguousIssue(len(outcomes), name="-".join(self.issue_names))
            if numeric
            else CategoricalIssue(values, name="-".join(self.issue_names))
        )
        return DiscreteCartesianOutcomeSpace(
            issues=(issue,),
            name=self.name,
        )

    def sample(
        self,
        n_outcomes: int,
        with_replacement: bool = False,
        fail_if_not_enough=True,
    ) -> Iterable[Outcome]:
        """
        Samples up to n_outcomes with or without replacement.

        This methor provides a base implementation that is not memory efficient.
        It will simply create a list of all outcomes using `enumerate()` and then
        samples from it. Specific outcome space types should override this method
        to improve its efficiency if possible.

        """
        outcomes = self.enumerate()
        outcomes = list(outcomes)
        if with_replacement:
            return random.choices(outcomes, k=n_outcomes)
        if fail_if_not_enough and n_outcomes > self.cardinality:
            raise ValueError("Cannot sample enough")
        random.shuffle(outcomes)
        return outcomes[:n_outcomes]

    def __iter__(self):
        return self.enumerate().__iter__()

    def __len__(self) -> int:
        return self.cardinality



"""
Genius Negotiator
An agent used to connect to GENIUS agents (ver 8.0.4) and allow them to join negotiation mechanisms

"""
from __future__ import annotations

import math
import os
import pathlib
import random
import tempfile

from negmas import warnings
from negmas.outcomes.base_issue import Issue

from ..common import MechanismState, NegotiatorMechanismInterface
from ..config import CONFIG_KEY_GENIUS_BRIDGE_JAR, NEGMAS_CONFIG
from ..inout import get_domain_issues
from ..negotiators import Controller
from ..outcomes import CartesianOutcomeSpace, Outcome, issues_to_xml_str
from ..preferences import UtilityFunction, make_discounted_ufun
from ..sao.common import ResponseType, SAOResponse, SAOState
from ..sao.negotiators import SAONegotiator
from .bridge import GeniusBridge
from .common import (
    DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT,
    DEFAULT_JAVA_PORT,
    get_free_tcp_port,
)
from .ginfo import AGENT_BASED_NEGOTIATORS, PARTY_BASED_NEGOTIATORS, TESTED_NEGOTIATORS

__all__ = [
    "GeniusNegotiator",
]

INTERNAL_SEP, ENTRY_SEP, FIELD_SEP = "<<s=s>>", "<<y,y>>", "<<sy>>"
FAILED, OK, TIMEOUT = "__FAILED__", "__OK__", "__TIMEOUT__"


class GeniusNegotiator(SAONegotiator):
    """
    Encapsulates a Genius Negotiator

    Args:
        preferences: The ufun of the negotiator [optional]
        name: Negotiator name [optional]
        parent: Parent `Controller`
        owner: The agent that owns the negotiator (if any)
        java_class_name: The java class name of the Geinus underlying agent
        domain_file_name: Optional domain file name (containing the negotiation issues or agenda)
        utility_file_name: Optional ufun file name (xml) from which a ufun will be loaded for the agent
        can_propose: The negotiator can propose
        auto_load_java: Load the genius bridge if needed
        port: The port to load the genius bridge to (or use if it is already loaded)
        genius_bridge_path: The path to the genius bridge
        strict: If True, raise exceptions if any exception is thrown by the agent or the bridge
                (or if the agent could not choose an action).
                If false, ignore these exceptions and assume a None return.
                If None use strict for n_steps limited negotiations and not strict for time_limit
                limited ones.
    """

    def __init__(
        self,
        preferences: UtilityFunction | None = None,
        name: str | None = None,
        parent: Controller | None = None,
        owner: Agent = None,  # type: ignore
        java_class_name: str | None = None,
        domain_file_name: str | pathlib.Path | None = None,
        utility_file_name: str | pathlib.Path | None = None,
        can_propose=True,
        auto_load_java: bool = True,
        port: int = DEFAULT_JAVA_PORT,
        genius_bridge_path: str | None = None,
        strict: bool | None = None,
    ):
        super().__init__(
            name=name,
            preferences=None,
            parent=parent,
            owner=owner,
        )
        self.__frozen_relative_time = None
        self.__destroyed = False
        self.__started = False
        self._strict = strict
        self.capabilities["propose"] = can_propose
        self.add_capabilities({"genius": True})
        self.genius_bridge_path = (
            genius_bridge_path
            if genius_bridge_path is not None
            else NEGMAS_CONFIG.get(
                CONFIG_KEY_GENIUS_BRIDGE_JAR, "~/negmas/files/geniusbridge.jar"
            )
        )
        self.java = None
        self.java_class_name = (
            java_class_name
            if java_class_name is not None
            else GeniusNegotiator.random_negotiator_name()
        )
        self._port = port
        self.domain_file_name = str(domain_file_name) if domain_file_name else None
        self.utility_file_name = str(utility_file_name) if utility_file_name else None
        self._my_last_offer = None
        self._preferences, self.discount = None, None
        self.issue_names = self.issues = self.issue_index = None
        self.auto_load_java = auto_load_java
        if domain_file_name is not None:
            # we keep original issues details so that we can create appropriate answers to Java
            self.issues = get_domain_issues(
                domain_file_name=domain_file_name,
            )
            if not self.issues:
                raise ValueError(f"Cannot read domain file {domain_file_name}")
            self.issue_names = [_.name for _ in self.issues]
            self.issue_index = dict(zip(self.issue_names, range(len(self.issue_names))))
        self.discount = None
        if utility_file_name is not None:
            self._preferences, self.discount = UtilityFunction.from_genius(
                utility_file_name,
                issues=self.issues,
            )
        # if ufun is not None:
        #     self._preferences = ufun
        self.base_utility = self._preferences
        self.__preferences_received = preferences
        self._temp_domain_file = self._temp_preferences_file = False
        self.connected = False

    @property
    def is_connected(self):
        return self.connected and self.java is not None

    @property
    def port(self):
        # if a port was not specified then we just set any random empty port to be used
        if not self.is_connected and self._port <= 0:
            self._port = get_free_tcp_port()
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @classmethod
    def robust_negotiators(cls) -> list[str]:
        """
        Returns a list of genius agents that were tested and seem to be robustly working with negmas
        """
        return TESTED_NEGOTIATORS

    @classmethod
    def negotiators(cls, agent_based=True, party_based=True) -> list[str]:
        """
        Returns a list of all available agents in genius 8.4.0

        Args:
            agent_based: Old agents based on the Java class Negotiator
            party_based: Newer agents based on the Java class AbstractNegotiationParty

        Returns:

        """
        r = []
        if agent_based:
            r += AGENT_BASED_NEGOTIATORS
        if party_based:
            r += PARTY_BASED_NEGOTIATORS
        return r

    @classmethod
    def random_negotiator_name(
        cls,
        agent_based=True,
        party_based=True,
    ):
        agent_names = cls.negotiators(agent_based=agent_based, party_based=party_based)
        return random.choice(agent_names)

    @classmethod
    def random_negotiator(
        cls,
        agent_based=True,
        party_based=True,
        port: int = DEFAULT_JAVA_PORT,
        domain_file_name: str | None = None,
        utility_file_name: str | None = None,
        auto_load_java: bool = False,
        can_propose=True,
        name: str | None = None,
    ) -> GeniusNegotiator:
        """
        Returns an agent with a random class name

        Args:
            name: negotiator name
            can_propose: Can this negotiator propose?
            auto_load_java: load the JVM if needed
            utility_file_name: Name of the utility xml file
            domain_file_name: Name of the domain XML file
            port: port number to use if the JVM is to be started
            agent_based: Old agents based on the Java class Negotiator
            party_based: Newer agents based on the Java class AbstractNegotiationParty


        Returns:
            GeniusNegotiator an agent with a random java class
        """
        agent_name = cls.random_negotiator_name(
            agent_based=agent_based, party_based=party_based
        )
        return GeniusNegotiator(
            java_class_name=agent_name,
            port=port,
            domain_file_name=domain_file_name,
            utility_file_name=utility_file_name,
            auto_load_java=auto_load_java,
            can_propose=can_propose,
            name=name,
        )

    def _create(self):
        """Creates the corresponding java agent"""
        if self.java is None:
            raise ValueError(f"Cannot create Genius Agent (no java instance)")
        aid = self.java.create_agent(self.java_class_name)  # type: ignore
        if aid == FAILED:
            raise ValueError(f"Cannot initialized {self.java_class_name}")
        return aid

    def _connect(self, path: str, port: int, auto_load_java: bool = False) -> bool:
        """
        Connects the negotiator to an appropriate genius-bridge running the actual agent
        """
        try:
            gateway = GeniusBridge.gateway(port)
        except:
            gateway = None
        if gateway is None:
            if auto_load_java:
                GeniusBridge.start(port=port, path=path)
            try:
                gateway = GeniusBridge.gateway(port)
            except:
                gateway = None
            if gateway == None:
                self.java = None
                return False
        self.java = gateway.entry_point  # type: ignore
        return True

    @property
    def java_name(self):
        if not self.java:
            return None
        return self.java.get_name(self.java_uuid)  # type: ignore

    def join(
        self,
        nmi: NegotiatorMechanismInterface,
        state: MechanismState,
        *,
        preferences: UtilityFunction | None = None,
        ufun: UtilityFunction | None = None,
        role: str = "negotiator",
    ) -> bool:
        if ufun:
            preferences = ufun
        if not preferences:
            preferences = self.__preferences_received
        result = super().join(
            nmi=nmi, state=state, preferences=preferences, ufun=None, role=role
        )
        if not result or self.ufun is None:
            return False
        # only connect to the JVM running genius-bridge if you are going to join a negotiation.
        if not self.is_connected:
            mechanism_port: int = nmi.params.get("genius_port", 0)
            if mechanism_port > 0:
                self.port = mechanism_port
            self.connected = self._connect(
                path=self.genius_bridge_path,
                port=self.port,
                auto_load_java=self.auto_load_java,
            )
            if not self.is_connected:
                return False

        self.java_uuid = self._create()

        if not isinstance(nmi.cartesian_outcome_space, CartesianOutcomeSpace):
            raise ValueError(
                f"Genius negotiators cannot be used with an out come space of type {nmi.cartesian_outcome_space.__class__.__name__}. Must use `CartesianOutcomeSpace`"
            )
        self.issues = nmi.cartesian_outcome_space.issues
        self.issue_names = [_.name for _ in self.issues]
        self.issue_index = dict(zip(self.issue_names, range(len(self.issue_names))))
        if nmi.cartesian_outcome_space is not None and self.domain_file_name is None:
            domain_file = tempfile.NamedTemporaryFile("w", suffix=".xml", delete=False)
            self.domain_file_name = domain_file.name
            domain_file.write(issues_to_xml_str(nmi.cartesian_outcome_space.issues))
            domain_file.close()
            self._temp_domain_file = True
        if preferences is not None and self.utility_file_name is None:
            utility_file = tempfile.NamedTemporaryFile("w", suffix=".xml", delete=False)
            self.utility_file_name = utility_file.name
            utility_file.write(
                UtilityFunction.to_xml_str(
                    preferences,
                    issues=nmi.cartesian_outcome_space.issues,
                    discount_factor=self.discount,
                )
            )
            utility_file.close()
            self._temp_preferences_file = True
        return result

    def _outcome2str(self, outcome):
        output = ""
        if not self.issue_names:
            nmi = self.nmi
            if not nmi:
                raise ValueError("Cannot send outcome to java without an NMI")
            self.issue_names = [_.name for _ in nmi.cartesian_outcome_space.issues]
        outcome_dict = dict(zip(self.issue_names, outcome))
        for i, v in outcome_dict.items():
            # todo check that the order here will be correct!!
            output += f"{i}{INTERNAL_SEP}{v}{ENTRY_SEP}"
        output = output[: -len(ENTRY_SEP)]
        return output

    def destroy_java_counterpart(self, state=None) -> None:
        if self.__started and not self.__destroyed:
            if self.java is not None:
                self.__frozen_relative_time = self.java.get_relative_time(  # type: ignore
                    self.java_uuid
                )
                result = self.java.on_negotiation_end(  # type: ignore
                    self.java_uuid,
                    None
                    if state is None
                    else self._outcome2str(state.agreement)
                    if state.agreement is not None
                    else None,
                )
                if result in (OK, TIMEOUT):
                    result = self.java.destroy_agent(self.java_uuid)  # type: ignore
                    if result in (FAILED, TIMEOUT) and self._strict:
                        raise ValueError(
                            f"{self._me()} ended the negotiation but failed to destroy the agent. A possible memory leak"
                        )
                elif self._strict:
                    raise ValueError(f"{self._me()} failed to end the negotiation!!")
            self.__destroyed = True
        if self._temp_preferences_file:
            try:
                os.unlink(self.utility_file_name)  # type: ignore
            except (FileNotFoundError, PermissionError):
                pass
            self._temp_preferences_file = False

        if self._temp_domain_file:
            try:
                os.unlink(self.domain_file_name)  # type: ignore
            except (FileNotFoundError, PermissionError):
                pass
            self._temp_domain_file = False

    def on_negotiation_end(self, state: MechanismState) -> None:
        """called when a negotiation is ended"""
        super().on_negotiation_end(state)
        self.destroy_java_counterpart(state)

    def on_negotiation_start(self, state: MechanismState) -> None:
        """Called when the info starts. Connects to the JVM."""
        super().on_negotiation_start(state=state)
        if self._strict is None and self.nmi is not None:
            self._strict = self.nmi.n_steps is not None and self.nmi.n_steps != float(
                "inf"
            )
        info = self.nmi
        if info is None:
            raise ValueError("Cannot start a negotiation without a NMI")
        if self._preferences is not None and self.utility_file_name is None:
            domain_file = tempfile.NamedTemporaryFile("w", suffix=".xml", delete=False)
            self.domain_file_name = domain_file.name
            domain_file.write(issues_to_xml_str(self._preferences.outcome_space.issues))  # type: ignore
            domain_file.close()
            self._temp_domain_file = True
            utility_file = tempfile.NamedTemporaryFile("w", suffix=".xml", delete=False)
            self.utility_file_name = utility_file.name
            if not isinstance(self.ufun, UtilityFunction):
                raise ValueError(
                    f"Genius Negotiator must have `UtilityFunctions` but you passed {self.ufun.__class__.__name__}"
                )
            utility_file.write(
                UtilityFunction.to_xml_str(
                    self.ufun,
                    issues=info.cartesian_outcome_space.issues,
                    discount_factor=self.discount,
                )
            )
            utility_file.close()
            self._temp_preferences_file = True
        if self.discount is not None and self.discount != 1.0 and self._preferences:
            if not isinstance(self.ufun, UtilityFunction):
                raise ValueError(
                    f"Genius Negotiator must have `UtilityFunctions` but you passed {self.ufun.__class__.__name__}"
                )
            self.set_preferences(
                make_discounted_ufun(
                    self.ufun,
                    discount_per_round=self.discount,
                    power_per_round=1.0,
                )
            )
        n_steps = -1 if info.n_steps is None else int(info.n_steps)  # number of steps
        n_seconds = (
            -1
            if info.time_limit is None or math.isinf(info.time_limit)
            else int(info.time_limit)
        )  # time limit
        timeout = (
            info.negotiator_time_limit
            if info.negotiator_time_limit and not math.isinf(info.negotiator_time_limit)
            else info.step_time_limit
            if info.step_time_limit and not math.isinf(info.step_time_limit)
            else info.time_limit
            if info.time_limit and not math.isinf(info.time_limit)
            else DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT
        )
        if timeout is None or math.isinf(timeout) or timeout < 0:
            timeout = DEFAULT_GENIUS_NEGOTIATOR_TIMEOUT

        if info.mechanism:
            timeout = min(timeout, info.mechanism._hidden_time_limit)

        if n_steps * n_seconds > 0:
            if n_steps < 0:
                if self._strict:
                    raise ValueError(
                        f"{self._me()}: Neither n_steps ({n_steps}) nor n_seconds ({n_seconds}) are given. Not allowed in strict mode"
                    )
                warnings.warn(
                    f"{self._me()}: Neither n_steps ({n_steps}) nor n_seconds ({n_seconds}) are given. This may lead to an infinite negotiation",
                    warnings.NegmasInfiniteNegotiationWarning,
                )
            else:
                # n_steps take precedence
                if self._strict:
                    raise ValueError(
                        f"{self._me()}: Both n_steps ({n_steps}) and n_seconds ({n_seconds}) are given. Not allowed in strict execution"
                    )
                warnings.warn(
                    f"{self._me()}: Both n_steps ({n_steps}) and n_seconds ({n_seconds}) are given. time_limit will be ignored",
                    warnings.NegmasStepAndTimeLimitWarning,
                )
                n_seconds, timeout = -1, -1
        timeout = int(timeout)
        try:
            result = self.java.on_negotiation_start(  # type: ignore
                self.java_uuid,  # java_uuid
                info.n_negotiators,  # number of agents
                n_steps,
                n_seconds,
                n_seconds > 0,
                self.domain_file_name,  # domain file
                self.utility_file_name,  # Negotiator file
                timeout,
                self._strict,
            )
            self.__started = result == OK
        except Exception as e:
            raise ValueError(
                f"{self._me()}: Cannot start negotiation: {str(e)}\nDomain file: {self.domain_file_name}\nUFun file: {self.domain_file_name}\n{'strict' if self._strict else 'non-strict'} {n_steps=} {timeout=}"
            )

    def cancel(self, reason=None) -> None:
        try:
            self.java.cancel(self.java_uuid)  # type: ignore
        except:
            pass

    @property
    def relative_time(self) -> float | None:
        if self.nmi is None or not self.nmi.state.started:
            return 0
        if self.nmi is not None and self.nmi.state.ended:
            return self.__frozen_relative_time
        t = self.java.get_relative_time(self.java_uuid)  # type: ignore
        if t < 0:
            if self._strict:
                raise ValueError(
                    f"{self._me()} cannot read relative time (returned {t})"
                )
            return None
        return t

    def _me(self):
        return (
            f"Agent {self.name} (jid: {self.java_uuid}) of type {self.java_class_name}"
        )

    def parse(self, action: str) -> tuple[ResponseType, Outcome | None]:
        """
        Parses an action into a ResponseType and an Outcome (if one is included)
        Args:
            action:

        Returns:

        """
        outcome = None
        if len(action) < 1:
            if self._strict:
                raise ValueError(f"{self._me()} received no actions while parsing")
            return ResponseType.REJECT_OFFER, None
        _, typ_, bid_str = action.split(FIELD_SEP)
        nmi = self.nmi
        if not nmi:
            raise ValueError("Cannot parse without an NMI")
        if typ_ == FAILED:
            raise ValueError(
                f"{self._me()} sent an action that cannot be parsed ({action})"
            )
        elif typ_ == TIMEOUT:
            if nmi.state.relative_time < 1.0:
                raise ValueError(
                    f"{self._me()} indicated that it timedout at relative time ({nmi.state.relative_time})"
                )

        issues = nmi.cartesian_outcome_space.issues

        def map_value(issue: Issue, val: str):
            if not issue.value_type:
                return val
            if issubclass(issue.value_type, tuple):
                return eval(val)
            return issue.value_type(val)  # type: ignore (It will always be a constructable type)

        if typ_ in ("Offer",) and (bid_str is not None and len(bid_str) > 0):
            try:
                values = {
                    _[0]: _[1]
                    for _ in [_.split(INTERNAL_SEP) for _ in bid_str.split(ENTRY_SEP)]
                }
                outcome = []
                for issue in issues:
                    outcome.append(map_value(issue, values[issue.name]))
                outcome = tuple(outcome)
            except Exception as e:
                if self._strict:
                    raise ValueError(
                        f"{self._me()} failed to parse {bid_str} of action {action} with exception {str(e)}"
                    )
                warnings.warn(
                    f"{self._me()} failed in parsing bid string: {bid_str} of action {action} with exception {str(e)}",
                    warnings.NegmasBrdigeParsingWarning,
                )

        if typ_ in (TIMEOUT, "Offer"):
            response = ResponseType.REJECT_OFFER
        elif typ_ == "Accept":
            response = ResponseType.ACCEPT_OFFER
        elif typ_ == "EndNegotiation":
            response = ResponseType.END_NEGOTIATION
        elif typ_ in ("NullOffer", "Failure", "NoAction"):
            if self._strict:
                raise ValueError(f"{self._me()} received {typ_} in action {action}")
            response = ResponseType.REJECT_OFFER
            outcome = None
        else:
            raise ValueError(
                f"{self._me()}: Unknown response: {typ_} in action {action}"
            )
        return response, tuple(outcome) if outcome else None

    def counter(self, state: SAOState, offer: Outcome | None):
        if offer is None and self._my_last_offer is not None and self._strict:
            raise ValueError(f"{self._me()} got counter with a None offer.")
        if offer is not None:
            received = self.java.receive_message(  # type: ignore
                self.java_uuid,
                state.current_proposer,
                "Offer",
                self._outcome2str(offer),
                state.step,
            )
            if self._strict and received == FAILED:
                raise ValueError(
                    f"{self._me()} failed to receive message in step {state.step}"
                )
        response, outcome = self.parse(
            self.java.choose_action(self.java_uuid, state.step)  # type: ignore
        )
        if response is None or (
            self._strict
            and (
                response
                not in (
                    ResponseType.REJECT_OFFER,
                    ResponseType.ACCEPT_OFFER,
                    ResponseType.END_NEGOTIATION,
                )
                or (response == ResponseType.REJECT_OFFER and outcome is None)
            )
        ):
            raise ValueError(
                f"{self._me()} returned a None counter offer in step {state.step}"
            )

        self._my_last_offer = outcome
        return SAOResponse(response, outcome)

    def propose(self, state):
        raise ValueError(
            f"{self._me()}: propose should never be called directly on GeniusNegotiator"
        )

    def __str__(self):
        name = super().__str__().split("/")
        return "/".join(name[:-1]) + f"/{self.java_class_name}/" + name[-1]



from __future__ import annotations

import numbers
import random
from abc import abstractmethod

import numpy as np

from negmas.helpers.numeric import get_one_float
from negmas.helpers.prob import Distribution, Real, ScipyDistribution
from negmas.outcomes import Outcome

from .base_ufun import BaseUtilityFunction, _General

__all__ = [
    "ProbUtilityFunction",
]


class ProbUtilityFunction(_General, BaseUtilityFunction):
    """A probablistic utility function. One that returns a probability distribution when called"""

    @abstractmethod
    def eval(self, offer: Outcome) -> Distribution:
        ...

    def to_prob(self) -> ProbUtilityFunction:
        return self

    @classmethod
    def generate_bilateral(
        cls,
        outcomes: int | list[Outcome],
        conflict_level: float = 0.5,
        conflict_delta=0.005,
        scale: float | tuple[float, float] = 0.5,
    ) -> tuple[ProbUtilityFunction, ProbUtilityFunction]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate.
                            1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.

        Examples:

            >>> from negmas.preferences import UtilityFunction
            >>> from negmas.preferences import conflict_level
            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.0
            ...                                             , conflict_delta=0.0)
            >>> print(conflict_level(u1, u2, outcomes=10))
            0.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=1.0
            ...                                             , conflict_delta=0.0)
            >>> print(conflict_level(u1, u2, outcomes=10))
            1.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.5
            ...                                             , conflict_delta=0.0)
            >>> 0.0 < conflict_level(u1, u2, outcomes=10) < 1.0
            True


        """
        from negmas.preferences.prob.mapping import ProbMappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        rand = np.random.random(n_outcomes)
        if conflict_level > 0.5:
            conflicting = 1.0 - u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = conflicting * conflict_level + rand * (1 - conflict_level)
        elif conflict_level < 0.5:
            same = u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = same * (1 - conflict_level) + rand * conflict_level
        else:
            u2 = rand

        # todo implement win_win correctly. Order the ufun then make outcomes with good outcome even better and vice
        # versa
        # u2 += u2 * win_win
        # u2 += np.random.random(n_outcomes) * conflict_delta
        u1 -= u1.min()
        u2 -= u2.min()
        u1 = u1 / u1.max()
        u2 = u2 / u2.max()
        if random.random() > 0.5:
            u1, u2 = u2, u1
        return (
            ProbMappingUtilityFunction(
                dict(
                    zip(
                        outcomes,
                        (
                            ScipyDistribution(
                                type="unifomr", loc=_, scale=get_one_float(scale)
                            )
                            for _ in u1
                        ),
                    )
                )
            ),
            ProbMappingUtilityFunction(
                dict(
                    zip(
                        outcomes,
                        (
                            ScipyDistribution(
                                type="unifomr", loc=_, scale=get_one_float(scale)
                            )
                            for _ in u2
                        ),
                    )
                )
            ),
        )

    @classmethod
    def generate_random_bilateral(
        cls, outcomes: int | list[Outcome], scale: float = 0.5
    ) -> tuple[ProbUtilityFunction, ProbUtilityFunction]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate. 1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.
            zero_summness: How zero-sum like are the two ufuns.


        """
        from negmas.preferences.prob.mapping import ProbMappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        u2 = np.random.random(n_outcomes)
        u1 -= u1.min()
        u2 -= u2.min()
        u1 /= u1.max()
        u2 /= u2.max()
        return (
            ProbMappingUtilityFunction(
                dict(
                    zip(
                        outcomes,
                        (
                            ScipyDistribution(
                                type="unifomr", loc=_, scale=get_one_float(scale)
                            )
                            for _ in u1
                        ),
                    )
                )
            ),
            ProbMappingUtilityFunction(
                dict(
                    zip(
                        outcomes,
                        (
                            ScipyDistribution(
                                type="unifomr", loc=_, scale=get_one_float(scale)
                            )
                            for _ in u2
                        ),
                    )
                )
            ),
        )

    @classmethod
    def generate_random(
        cls,
        n: int,
        outcomes: int | list[Outcome],
        normalized: bool = True,
        scale: float | tuple[float, float] = 0.5,
    ) -> list[ProbUtilityFunction]:
        """Generates N mapping utility functions

        Args:
            n: number of utility functions to generate
            outcomes: number of outcomes to use
            normalized: if true, the resulting ufuns will be normlized between zero and one.


        """
        from negmas.preferences.prob.mapping import ProbMappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        else:
            outcomes = list(outcomes)
        n_outcomes = len(outcomes)
        ufuns = []
        for _ in range(n):
            u1 = np.random.random(n_outcomes)
            if normalized:
                u1 -= u1.min()
                u1 /= u1.max()
            ufuns.append(
                ProbMappingUtilityFunction(
                    dict(
                        zip(
                            outcomes,
                            (
                                ScipyDistribution(
                                    type="unifomr", loc=_, scale=get_one_float(scale)
                                )
                                for _ in u1
                            ),
                        )
                    )
                )
            )
        return ufuns

    def __call__(self, offer: Outcome | None) -> Distribution:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:

            - It calls the abstract method `eval` after opationally adjusting the
              outcome type.
            - It is preferred to override eval instead of directly overriding this method
            - You cannot return None from overriden eval() functions but raise an exception (ValueError) if it was
              not possible to calculate the Value.
            - Return a float from your `eval` implementation.
            - Return the reserved value if the offer was None

        Returns:
            The utility of the given outcome
        """
        if offer is None:
            return ScipyDistribution("uniform", loc=self.reserved_value, scale=0.0)
        v = self.eval(offer)
        if isinstance(v, float):
            return Real(v)
        return v


class ProbAdapter(ProbUtilityFunction):
    """
    Adapts any utility function to act as a probabilistic utility function (i.e. returning a `Distribution` )
    """

    def __init__(self, ufun: BaseUtilityFunction):
        self._ufun = ufun

    def eval(self, offer):
        v = self._ufun.eval(offer)
        if isinstance(v, numbers.Real):
            return Real(v)
        return v

    def to_stationary(self):
        return ProbAdapter(self._ufun.to_stationary())



#!/usr/bin/env python
"""
A set of utilities to handle strings
"""
from __future__ import annotations

import datetime
import itertools
import math
import random
import re
import socket
import string
import traceback
from collections import defaultdict

__all__ = [
    "shortest_unique_names",
    "snake_case",
    "camel_case",
    "unique_name",
    "pretty_string",
    "exception2str",
    "humanize_time",
    "shorten",
]

COMMON_NAME_PARTS = (
    "Mechanism",
    "Negotiator",
    "Agent",
    "Controller",
    "Acceptance",
    "Component",
    "Model",
    "Strategy",
    "Offering",
    "Entity",
)
"""Default parts of names removed by `shorten` """


def shorten(name: str, length: int = 4, common_parts=COMMON_NAME_PARTS) -> str:
    """
    Returns a short version of the name.

    Remarks:
        - Removes common parts of names in negmas like Negotiator, Agent, etc
        - Keeps Capital letters up to the given length
        - Adds some of the lowercase letters to fit the length
        - If the input is shorter than the length, it is returned as it is
    """
    for p in common_parts:
        if len(name) <= len(p):
            continue
        name = name.replace(p, "")
    if len(name) <= length:
        return name
    caps = [_ for _ in name if _.isupper()]
    if len(caps) >= length:
        return "".join(caps[:length])
    needed = length - len(caps)
    caps = []
    for c in name:
        if len(caps) >= length:
            break
        if c.isupper():
            caps.append(c)
            continue
        if needed < 1:
            continue
        needed -= 1
        caps.append(c)
    return "".join(caps[:length])


def unique_name(
    base,
    add_time=True,
    add_host=False,
    rand_digits=8,
    sep="/",
) -> str:
    """Return a unique name.

    Can be used to return a unique directory name on the givn base.

    Args:
        base: (any): base path/string (it is converted to string whatever it is)
        add_time (bool, optional): Defaults to True. Add current time
        rand_digits (int, optional): Defaults to 8. The number of random
            characters to add to the name

    Examples:

        >>> a = unique_name('')
        >>> len(a) == 8 + 1 + 6 + 8 + 6
        True

    Returns:
        str: The unique name.

    """
    base = str(base)
    _time, rand_part = "", ""
    host_part = socket.gethostname() if add_host else ""
    if rand_digits > 0:
        rand_part = "".join(
            random.choices(string.digits + string.ascii_letters, k=rand_digits)
        )
    if add_time:
        _time = datetime.datetime.now().strftime("%Y%m%dH%H%M%S%f")
    sub = _time + host_part + rand_part
    if len(sub) == 0:
        return base
    if len(base) == 0:
        return sub
    return f"{str(base)}{sep}{sub}"


def shortest_unique_names(
    strs: list[str], sep=".", max_compression=False, guarantee_unique=False
):
    """
    Finds the shortest unique strings starting from the end of each input
    string based on the separator.

    The final strings will only be unique if the inputs are unique.

    Args:
        strs: A list of strings
        sep: The separator used to separate fields in each string
        max_compression: If True, each string will be further compressed
                         by taking the shortest prefix that keeps the
                         strings unique (if they were originally unique)
        guarantee_unique: If given, random characters will be postfixed on
                         strings to guarantee uniquness

    Example:
        given ["a.b.cat", "d.e.f", "a.d.cat"] it will generate ["b.c", "f", "d.cat"]
        if max_compression was false and will generate ["b", "f", "d"] if it was
        True
    """
    if len(strs) < 2:
        return strs
    if guarantee_unique and len(set(strs)) != len(strs):
        chars = string.digits + string.ascii_letters
        for i in range(len(strs) - 1):
            others = set(strs[:i] + strs[i + 1 :])
            while strs[i] in others:
                for a in chars:
                    if strs[i] + a not in others:
                        strs[i] = strs[i] + a
                        break
                else:
                    strs[i] = strs[i] + unique_name("", False, False, 1, "")

    lsts = [_.split(sep) for _ in strs]
    names = [_[-1] for _ in lsts]
    if len(names) != len(set(names)):
        locs = defaultdict(list)
        for i, s in enumerate(names):
            locs[s].append(i)
        mapping = {"": ""}
        for s, l in locs.items():
            if len(s) < 1:
                continue
            if len(l) == 1:
                mapping[strs[l[0]]] = s
                continue
            strs_new = [sep.join(lsts[_][:-1]) for _ in l]
            prefixes = shortest_unique_names(
                strs_new, sep, max_compression, guarantee_unique
            )
            for loc, prefix in zip(l, prefixes):
                x = sep.join([prefix, s])
                if x.startswith(sep):
                    x = x[len(sep) :]
                mapping[strs[loc]] = x
        strs = [mapping[_] for _ in strs]
    else:
        strs = names
    if not max_compression:
        return strs
    for i, s in enumerate(strs):
        for j in range(1, len(s)):
            for k in itertools.chain(range(i), range(i + 1, len(strs))):
                if strs[k][:j] == s[:j]:
                    break
            else:
                strs[i] = s[:j]
                break
    return strs


def snake_case(s: str) -> str:
    """Converts a string from CamelCase to snake_case

    Example:

        >>> print(snake_case('ThisIsATest'))
        this_is_a_test



    Args:
        s: input string

    Returns:
        str: converted string
    """
    return (
        re.sub("(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))", "_\\1", s).lower().strip("_")
    )


def camel_case(
    s: str, capitalize_first: bool = False, lower_first: bool = False
) -> str:
    """Converts a string from snake_case to CamelCase

    Example:

        >>> print(camel_case('this_is_a_test'))
        thisIsATest
        >>> print(camel_case('this_is_a_test', capitalize_first=True))
        ThisIsATest
        >>> print(camel_case('This_is_a_test', lower_first=True))
        thisIsATest
        >>> print(camel_case('This_is_a_test'))
        ThisIsATest

    Args:
        s: input string
        capitalize_first: if true, the first character will be capitalized
        lower_first: If true, the first character will be lowered

    Returns:
        str: converted string
    """
    if len(s) < 1:
        return s
    parts = s.split("_")
    if capitalize_first:
        parts = [_.capitalize() for _ in parts]
    elif lower_first:
        parts = [parts[0].lower()] + [_.capitalize() for _ in parts[1:]]
    else:
        parts = [parts[0]] + [_.capitalize() for _ in parts[1:]]

    return "".join(parts)


def _pretty_string(src, dpth=0, current_key="", tab_size=2) -> str:
    """Recursively print nested elements.

    Args:
        dpth (int): Current depth
        current_key (str): Current key being printed
        tab_size: Tab size in spaces

    Returns:
        str: The pretty version of the input
    """

    def tabs(n):
        return " " * n * tab_size  # or 2 or 8 or...

    output = ""
    if isinstance(src, dict):
        output += tabs(dpth) + "{\n"
        for key, value in src.items():
            output += _pretty_string(value, dpth + 1, key) + "\n"
        output += tabs(dpth) + "}"
    elif isinstance(src, list) or isinstance(src, tuple):
        output += tabs(dpth) + "[\n"
        for litem in src:
            output += _pretty_string(litem, dpth + 1) + "\n"
        output += tabs(dpth) + "]"
    else:
        if len(current_key) > 0:
            output += tabs(dpth) + f'"{current_key}":{src}'
        else:
            output += tabs(dpth) + "%s" % src
    return output


def pretty_string(src, tab_size=2, compact=False) -> str:
    """Recursively print nested elements.

    Args:
        src (Any): The source to be converted to a printable string
        tab_size (int): Tab size in spaces
        compact (bool): If true the output is  converted into a single line

    Returns:
        str: The pretty version of the input

    Remarks:
        - This function assumes that the patterns `` "`` and ``":`` do not appear anywhere in the input.
          If they appear, the space, : will be removed.
    """
    s = _pretty_string(src, dpth=0, current_key="", tab_size=tab_size)
    if compact:
        return s.replace("\n", "")

    else:
        return s.replace(' "', " ").replace('":', ":")


def exception2str(limit=None, chain=True) -> str:
    return traceback.format_exc(limit=limit, chain=chain)


def humanize_time(
    secs,
    align=False,
    always_show_all_units=False,
    show_us=False,
    show_ms=False,
    always_show_from="",
):
    """
    Prints time that is given as seconds in human readable form. Useful only for times >=1sec.

    :param secs: float: number of seconds
    :param align: bool, optional: whether to align outputs so that they all take the same size (not implemented)
    :param always_show_all_units: bool, optional: Whether to always show days, hours, and minutes even when they
                                are zeros. default False
    :param always_show_from: One of d,h,m,s,ms,u (day, hour, minute, second, milli-sec, micro-sec) to always show
                             as well as everything shorter than it (i.e passing 'm' shows minutes, seconds, ... etc)
    :param show_us: bool, if given microseconds and milliseconds will be shown
    :param show_ms: bool, if given milliseconds will be shown
    :return: str: formated string with the humanized form
    """
    if show_us:
        secs *= 1_000_000
        units = [
            ("d", 86400_000_000),
            ("h", 3600_000_000),
            ("m", 60_000_000),
            ("s", 1_000_000),
            ("ms", 1000),
            ("u", 1),
        ]
    elif show_ms:
        secs *= 1_000
        units = [
            ("d", 86400_000),
            ("h", 3600_000),
            ("m", 60_000),
            ("s", 1_000),
            ("ms", 1),
        ]
    else:
        units = [("d", 86400), ("h", 3600), ("m", 60), ("s", 1)]
    parts = []
    for unit, mul in units:
        if unit == always_show_from:
            always_show_all_units = True
        if secs / mul >= 1 or mul == 1 or always_show_all_units:
            if mul > 1:
                n = int(math.floor(secs / mul))
                secs -= n * mul
            else:
                n = secs if secs != int(secs) else int(secs)
            if align:
                parts.append("%2d%s%s" % (n, unit, ""))
            else:
                parts.append("%2d%s%s" % (n, unit, ""))
    return "".join(parts)



from __future__ import annotations

from abc import abstractmethod
from collections import namedtuple
from typing import TYPE_CHECKING

from attr import define, field

from ...negotiators.components.component import Component

if TYPE_CHECKING:
    from negmas import ResponseType
    from negmas.outcomes import Outcome
    from negmas.sao import SAONegotiator, SAOState

__all__ = [
    "SAOComponent",
    "AcceptanceStrategy",
    "OfferingStrategy",
    "ProposalStrategy",
    "Model",
]


@define
class SAOComponent(Component):
    _negotiator: SAONegotiator | None = field(default=None, kw_only=True)

    def before_proposing(self, state: SAOState):
        """
        Called before proposing
        """

    def after_proposing(self, state: SAOState, offer: Outcome | None):
        """
        Called after proposing
        """

    def before_responding(self, state: SAOState, offer: Outcome | None):
        """
        Called before offering
        """

    def after_responding(
        self, state: SAOState, offer: Outcome | None, response: ResponseType
    ):
        """
        Called before offering
        """

    def on_partner_joined(self, partner: str):
        """
        Called when a partner joins the negotiation.

        This is only receivd if the mechanism is sending notifications.
        """

    def on_partner_left(self, partner: str):
        """
        Called when a partner leaves the negotiation.

        This is only receivd if the mechanism is sending notifications.
        """

    def on_partner_ended(self, partner: str):
        """
        Called when a partner ends the negotiation.

        Note  that the negotiator owning this component may never receive this
        offer.
        This is only receivd if the mechanism is sending notifications
        on every offer.
        """

    def on_partner_proposal(
        self, state: SAOState, partner_id: str, offer: Outcome
    ) -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `MechanismState` giving the state of the negotiation when the offer was porposed.
            partner_id: The ID of the agent who proposed
            offer: The proposal.

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_refused_to_propose(self, state: SAOState, partner_id: str) -> None:
        """
        A callback called by the mechanism when a partner refuses to propose

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner refused to offer.
            partner_id: The ID of the agent who refused to propose

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_response(
        self,
        state: SAOState,
        partner_id: str,
        outcome: Outcome | None,
        response: ResponseType,
    ) -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner responded.
            partner_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """


@define
class AcceptanceStrategy(SAOComponent):
    def respond(self, state: SAOState, offer: Outcome | None) -> ResponseType:
        """Called to respond to an offer. This is the method that should be overriden to provide an acceptance strategy.

        Args:
            state: a `SAOState` giving current state of the negotiation.
            offer: offer being tested

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        return self(state, offer)

    def __not__(self):
        return RejectionStrategy(self)

    @abstractmethod
    def __call__(self, state: SAOState, offer: Outcome | None) -> ResponseType:
        return self.respond(state, offer)

    def __and__(self, s: AcceptanceStrategy):
        from .acceptance import AllAcceptanceStrategies

        if self.negotiator != s.negotiator:
            raise ValueError(f"Cannot combine strategies with different negotiators")
        return AllAcceptanceStrategies([self, s])

    def __or__(self, s: AcceptanceStrategy):
        from .acceptance import AnyAcceptanceStrategy

        if self.negotiator != s.negotiator:
            raise ValueError(f"Cannot combine strategies with different negotiators")
        return AnyAcceptanceStrategy([self, s])


@define
class RejectionStrategy(AcceptanceStrategy):
    """
    Reverses the decision of an acceptance strategy (Rejects if the original was accepting and Accepts in any other case)
    """

    a: AcceptanceStrategy

    def __call__(self, state: SAOState, offer: Outcome | None) -> ResponseType:
        response = self.a(state, offer)
        if response == ResponseType.ACCEPT_OFFER:
            return ResponseType.REJECT_OFFER
        return ResponseType.ACCEPT_OFFER


@define
class OfferingStrategy(SAOComponent):
    _current_offer: tuple[int, Outcome | None] = field(init=False, default=(-1, None))

    def propose(self, state: SAOState) -> Outcome | None:
        """Propose an offer or None to refuse.

        Args:
            state: `SAOState` giving current state of the negotiation.

        Returns:
            The outcome being proposed or None to refuse to propose

        Remarks:
            - This function guarantees that no agents can propose something with a utility value

        """
        if self._current_offer[0] != state.step:
            offer = self(state)
            self._current_offer = (state.step, offer)
        return self._current_offer[1]

    @abstractmethod
    def __call__(self, state: SAOState) -> Outcome | None:
        ...

    def __and__(self, s: OfferingStrategy):
        from .offering import UnanimousConcensusOfferingStrategy

        if self.negotiator != s.negotiator:
            raise ValueError(f"Cannot combine strategies with different negotiators")
        return UnanimousConcensusOfferingStrategy([self, s])

    def __or__(self, s: OfferingStrategy):
        from .offering import RandomConcensusOfferingStrategy

        if self.negotiator != s.negotiator:
            raise ValueError(f"Cannot combine strategies with different negotiators")
        return RandomConcensusOfferingStrategy([self, s])


ProposalStrategy = OfferingStrategy
"""An alias for `OfferingStrategy` """

Model = SAOComponent
"""An alias for `SAOComponent` """


FilterResult = namedtuple("FilterResult", ["next", "save"])



from __future__ import annotations

import math
import random
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from attr import define, field

from negmas import warnings
from negmas.common import PreferencesChangeType, Value

from .base import FilterResult, OfferingStrategy
from .concession import ConcessionRecommender
from .models.ufun import UFunModel

if TYPE_CHECKING:
    from negmas.common import PreferencesChange
    from negmas.outcomes import Outcome
    from negmas.sao import SAOState
    from negmas.sao.negotiators.base import SAONegotiator


__all__ = [
    "LimitedOutcomesOfferingStrategy",
    "NegotiatorOfferingStrategy",
    "ConcensusOfferingStrategy",
    "RandomConcensusOfferingStrategy",
    "UnanimousConcensusOfferingStrategy",
    "UtilBasedConcensusOfferingStrategy",
    "MyBestConcensusOfferingStrategy",
    "MyWorstConcensusOfferingStrategy",
    "NoneOfferingStrategy",
    "RandomOfferingStrategy",
    "OfferTop",
    "OfferBest",
    "TFTOfferingStrategy",
]


@define
class TFTOfferingStrategy(OfferingStrategy):
    """
    An acceptance strategy that concedes as much as the partner (or more)
    """

    partner_ufun: UFunModel
    recommender: ConcessionRecommender
    stochastic: bool = False
    _partner_offer: Outcome | None = field(init=False, default=None)

    def before_responding(self, state: SAOState, offer: Outcome | None):
        self._partner_offer = offer

    def __call__(self, state):
        if not self.negotiator or not self.negotiator.ufun:
            return None
        partner_u = (
            float(self.partner_ufun.eval_normalized(self._partner_offer, True))
            if self._partner_offer
            else 1.0
        )
        partner_concession = 1.0 - partner_u
        my_concession = self.recommender(partner_concession, state)
        if not math.isfinite(my_concession):
            warnings.warn(
                f"Got {my_concession} for concession which is unacceptable. Will use no concession"
            )
            my_concession = 0.0
        if not (-1e-6 <= my_concession <= 1.0000001):
            warnings.warn(f"{my_concession} is negative or above 1")
            my_concession = 0.0
        target_utility = 1.0 - float(my_concession)
        if self.stochastic:
            return self.negotiator.ufun.invert().one_in(
                (target_utility, 1.0), normalized=True
            )
        return self.negotiator.ufun.invert().worst_in(
            (target_utility, 1.0), normalized=True
        )


@define
class OfferBest(OfferingStrategy):
    """
    Offers Only the best outcome.

    Remarks:
        - You can pass the  best outcome if you know it as `best` otherwise it will find it.
    """

    _best: Outcome | None = None

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            return
        _, self._best = self.negotiator.ufun.extreme_outcomes()

    def __call__(self, state: SAOState) -> Outcome | None:
        return self._best


@define
class OfferTop(OfferingStrategy):
    """
    Offers outcomes that are in the given top fraction or top `k`. If neither is given it reverts to only offering the best outcome

    Remarks:
        - The outcome-space is always discretized and the constraints `fraction` and `k` are applied to the discretized space
    """

    fraction: float = 0.0
    k: int = 1

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            return
        if any(
            _.type
            not in (
                PreferencesChangeType.Scaled,
                PreferencesChangeType.ReservedOutcome,
                PreferencesChangeType.ReservedValue,
            )
            for _ in changes
        ):
            self.negotiator.ufun.invert().init()

    def __call__(self, state: SAOState) -> Outcome | None:
        if not self.negotiator or not self.negotiator.ufun:
            return None
        top_k = self.negotiator.ufun.invert().within_indices((0, self.k))
        top_f = self.negotiator.ufun.invert().within_fractions((0.0, self.fraction))
        top = list(set(top_k + top_f))
        return random.choice(top)


@define
class NoneOfferingStrategy(OfferingStrategy):
    """
    Always offers `None` which means it never gets an agreement.
    """

    def __call__(self, state: SAOState) -> Outcome | None:
        return None


@define
class RandomOfferingStrategy(OfferingStrategy):
    """
    Always offers `None` which means it never gets an agreement.
    """

    def __call__(self, state: SAOState) -> Outcome | None:
        if not self.negotiator or not self.negotiator.nmi:
            return None
        return self.negotiator.nmi.random_outcomes(1)[0]


@define
class LimitedOutcomesOfferingStrategy(OfferingStrategy):
    """
    Offers from a given list of outcomes
    """

    outcomes: list[Outcome] | None
    prob: list[float] | None = None
    p_ending: float = 0.0

    def __call__(self, state: SAOState, retry=False) -> Outcome | None:
        if not self.negotiator or not self.negotiator.nmi:
            return None
        if random.random() < self.p_ending - 1e-7:
            return None
        if not self.prob or not self.outcomes:
            return random.choice(
                self.outcomes
                if self.outcomes
                else list(self.negotiator.nmi.discrete_outcomes())
            )
        r, s = random.random(), 0.0
        for w, p in zip(self.outcomes, self.prob):
            s += p
            if r <= s:
                return w
        if retry:
            return None
        if s > 0.999:
            return self.outcomes[-1]
        self.prob = [_ / s for _ in self.prob]
        return self.propose(state)


@define
class NegotiatorOfferingStrategy(OfferingStrategy):
    """
    Uses a negotiator as an offering strategy
    """

    proposer: SAONegotiator = field(kw_only=True)

    def __call__(self, state: SAOState) -> Outcome | None:
        return self.proposer.propose(state)


@define
class ConcensusOfferingStrategy(OfferingStrategy, ABC):
    """
    Offers based on concensus of multiple strategies
    """

    strategies: list[OfferingStrategy]

    def filter(self, indx: int, offer: Outcome | None) -> FilterResult:
        """
        Called with the decision of each strategy in order.


        Remarks:
            - Two decisions need to be made:

              1. Should we continue trying other strategies
              2. Should we save this result.
        """
        return FilterResult(True, True)

    @abstractmethod
    def decide(
        self, indices: list[int], responses: list[Outcome | None]
    ) -> Outcome | None:
        """
        Called to make a final decsision given the decisions of the stratgeis with indices `indices` (see `filter` for filtering rules)
        """

    def __call__(self, state: SAOState) -> Outcome | None:
        selected, selected_indices = [], []
        for i, s in enumerate(self.strategies):
            response = s.propose(state)
            r = self.filter(i, response)
            if not r.next:
                break
            if r.save:
                selected.append(response)
                selected_indices.append(i)

        return self.decide(selected_indices, selected)


@define
class UnanimousConcensusOfferingStrategy(ConcensusOfferingStrategy):
    """
    Offers only if all offering strategies gave exactly the same outcome
    """

    def decide(
        self, indices: list[int], responses: list[Outcome | None]
    ) -> Outcome | None:
        outcomes = set(responses)
        if len(outcomes) != 1:
            return None
        return list(outcomes)[0]


@define
class RandomConcensusOfferingStrategy(ConcensusOfferingStrategy):
    """
    Offers a random response from the list of stratgies (different strategy every time).
    """

    prob: list[float] | None = None

    def __attrs_post_init__(self):
        if not self.prob:
            return
        s = sum(self.prob)
        self.prob = [_ / s for _ in self.prob]

    def decide(
        self, indices: list[int], responses: list[Outcome | None]
    ) -> Outcome | None:
        if not self.prob:
            return random.choice(responses)

        r, s = random.random(), 0.0
        for i, p in enumerate(self.prob):
            s += p
            if r <= s:
                return responses[i]
        if s > 0.999:
            return responses[-1]
        raise ValueError(f"sum of probabilities is less than 1: {s}")


@define
class UtilBasedConcensusOfferingStrategy(ConcensusOfferingStrategy, ABC):
    """
    Offers from the list of stratgies (different strategy every time) based on outcome utilities
    """

    @abstractmethod
    def decide_util(self, utils: list[Value]) -> int:
        """
        Returns the index to chose based on utils
        """

    def decide(
        self, indices: list[int], responses: list[Outcome | None]
    ) -> Outcome | None:
        if not self.negotiator.ufun:
            raise ValueError(f"Cannot decide because I have no ufun")
        return responses[
            self.decide_util([self.negotiator.ufun(_) for _ in set(responses)])
        ]


@define
class MyBestConcensusOfferingStrategy(UtilBasedConcensusOfferingStrategy):
    """
    Offers my best outcome from the list of stratgies (different strategy every time).
    """

    def decide_util(self, utils: list[Value]) -> int:
        return max(range(len(utils)), key=lambda x: utils[x])


@define
class MyWorstConcensusOfferingStrategy(UtilBasedConcensusOfferingStrategy):
    """
    Offers my worst outcome from the list of stratgies (different strategy every time) based on outcome utilities
    """

    def decide_util(self, utils: list[Value]) -> int:
        return min(range(len(utils)), key=lambda x: utils[x])



from __future__ import annotations

import pprint
from typing import TYPE_CHECKING, Iterable

import numpy as np

from negmas.generics import iget, ikeys, ivalues
from negmas.helpers.prob import ScipyDistribution
from negmas.outcomes import Issue, Outcome

from ...helpers.prob import (
    Distribution,
    ScipyDistribution,
    make_distribution,
    uniform_around,
)
from ..crisp.mapping import MappingUtilityFunction
from ..mixins import StationaryMixin
from ..prob_ufun import ProbUtilityFunction
from .mapping import ProbMappingUtilityFunction

if TYPE_CHECKING:
    from ..base import Value

__all__ = ["IPUtilityFunction"]


class IPUtilityFunction(StationaryMixin, ProbUtilityFunction):
    """
    Independent Probabilistic Utility Function.

    Args:

        outcomes: Iterable of outcomes
        distributions: the distributions. One for each outcome
        name: ufun name

    Examples:

        >>> outcomes = [('o1',), ('o2',)]
        >>> f = IPUtilityFunction(outcomes=outcomes
        ...         , distributions=[ScipyDistribution(type='uniform', loc=0.0, scale=0.5)
        ...         , ScipyDistribution(type='uniform', loc=0.1, scale=0.5)])
        >>> str(f(('o1',)))
        'U(0.0, 0.5)'

    """

    def __init__(
        self,
        outcomes: Iterable[Outcome],
        distributions: Iterable[Distribution] | None = None,
        issue_names: Iterable[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        outcomes = list(outcomes)
        distributions = (
            list(distributions)
            if distributions is not None
            else [
                ScipyDistribution(type="uniform", loc=0.0, scale=1.0)
                for _ in range(len(outcomes))
            ]
        )

        if len(outcomes) < 1:
            raise ValueError(
                "IPUtilityFunction cannot be initialized with zero outcomes"
            )
        self.tupelized = False

        self.n_issues = len(outcomes[0])
        if issue_names is None:
            self.issue_names = sorted(ikeys(outcomes[0]))
        else:
            self.issue_names = range(len(outcomes[0]))

        self.issue_keys = dict(zip(range(self.n_issues), self.issue_names))

        if not isinstance(outcomes[0], tuple):
            outcomes = [
                tuple(iget(_, key, None) for key in self.issue_names) for _ in outcomes
            ]
            self.tupelized = True
        self.distributions = dict(zip(outcomes, distributions))

    @classmethod
    def random(
        cls,
        outcome_space,
        reserved_value,
        normalized=True,
        dist_limits=(0.0, 1.0),
        **kwargs,
    ) -> IPUtilityFunction:
        """Generates a random ufun of the given type"""
        raise NotImplementedError("random hyper-rectangle ufuns are not implemented")

    def key(self, outcome: Outcome):
        """
        Returns the key of the given outcome in self.distributions.

        Args:
            outcome:

        Returns:
            tuple

        """
        return outcome

    def distribution(self, outcome: Outcome) -> Distribution:
        """
        Returns the distributon associated with a specific outcome
        Args:
            outcome:

        Returns:

        """
        return self.distributions[self.key(outcome)]

    @classmethod
    def from_mapping(
        cls,
        mapping: dict[Outcome, Value],
        range: tuple[float, float] = (0.0, 1.0),
        uncertainty: float = 0.5,
        variability: float = 0.0,
        reserved_value: float = float("-inf"),
    ) -> IPUtilityFunction:
        """
        Generates a distribution from which `u` may have been sampled
        Args:
            mapping: mapping from outcomes to float values
            range: range of the utility_function values
            uncertainty: uncertainty level
            variability: The variability within the ufun
            reserved_value: The reserved value

        Examples:

            - No uncertainty
            >>> mapping = dict(zip([('o1',), ('o2',)], [0.3, 0.7]))
            >>> p = IPUtilityFunction.from_mapping(mapping, uncertainty=0.0)
            >>> print(p)
            {('o1',): 0.3, ('o2',): 0.7}

            - Full uncertainty
            >>> mapping=dict(zip([('o1',), ('o2',)], [0.3, 0.7]))
            >>> p = IPUtilityFunction.from_mapping(mapping, uncertainty=1.0)
            >>> print(p)
            {('o1',): U(0.0, 1.0), ('o2',): U(0.0, 1.0)}

            - some uncertainty
            >>> mapping=dict(zip([('o1',), ('o2',)], [0.3, 0.7]))
            >>> p = IPUtilityFunction.from_mapping(mapping, uncertainty=0.1)
            >>> print([_.scale for _ in p.distributions.values()])
            [0.1, 0.1]
            >>> for k, v in p.distributions.items():
            ...     assert v.loc <= mapping[k]

        Returns:
            a new IPUtilityFunction
        """
        outcomes = list(mapping.keys())
        if isinstance(uncertainty, Iterable):
            uncertainties = uncertainty
        elif variability <= 0.0:
            uncertainties = [uncertainty] * len(outcomes)
        else:
            uncertainties = (
                uncertainty
                + (np.random.rand(len(outcomes)) - 0.5) * variability * uncertainty
            ).tolist()

        return IPUtilityFunction(
            outcomes=outcomes,
            distributions=[
                uniform_around(value=float(mapping[o]), uncertainty=u, range=range)
                for o, u in zip(outcomes, uncertainties)
            ],
            reserved_value=reserved_value,
        )

    @classmethod
    def from_preferences(
        cls,
        u: ProbMappingUtilityFunction,
        range: tuple[float, float] = (0.0, 1.0),
        uncertainty: float = 0.5,
        variability: float = 0.0,
    ) -> IPUtilityFunction:
        """
        Generates a distribution from which `u` may have been sampled
        Args:
            u:
            range: range of the utility_function values
            uncertainty: uncertainty level

        Examples:

            - No uncertainty
            >>> u = MappingUtilityFunction(mapping=dict(zip([('o1',), ('o2',)], [0.3, 0.7])))
            >>> p = IPUtilityFunction.from_preferences(u, uncertainty=0.0)
            >>> print(p)
            {('o1',): 0.3, ('o2',): 0.7}

            - Full uncertainty
            >>> u = MappingUtilityFunction(mapping=dict(zip([('o1',), ('o2',)], [0.3, 0.7])))
            >>> p = IPUtilityFunction.from_preferences(u, uncertainty=1.0)
            >>> print(p)
            {('o1',): U(0.0, 1.0), ('o2',): U(0.0, 1.0)}

            - some uncertainty
            >>> u = MappingUtilityFunction(mapping=dict(zip([('o1',), ('o2',)], [0.3, 0.7])))
            >>> p = IPUtilityFunction.from_preferences(u, uncertainty=0.1)
            >>> print([_.scale for _ in p.distributions.values()])
            [0.1, 0.1]
            >>> for k, v in p.distributions.items():
            ...     assert v.loc <= u(k)


        Returns:
            a new IPUtilityFunction
        """
        if isinstance(u.mapping, dict):
            return cls.from_mapping(
                u.mapping,
                range=range,
                uncertainty=uncertainty,
                variability=variability,
                reserved_value=u.reserved_value,
            )
        if not u.outcome_space:
            raise ValueError("Unknown outcome space")
        if not u.outcome_space.is_discrete():
            raise ValueError(
                "Cannot be constructed from a ufun with an infinite outcome space"
            )
        outcomes = u.outcome_space.enumerate()  # type: ignore (I know that it is a discrete space)
        d = dict(zip(outcomes, (u(_) for _ in outcomes)))
        return cls.from_mapping(
            d,  # type: ignore
            range=range,
            uncertainty=uncertainty,
            variability=variability,
            reserved_value=u.reserved_value,
        )

    def is_state_dependent(self):
        return True

    def sample(self) -> MappingUtilityFunction:
        """
        Samples the utility_function distribution to create a mapping utility function


        Examples:
            >>> import random
            >>> f = IPUtilityFunction(outcomes=[('o1',), ('o2',)]
            ...         , distributions=[ScipyDistribution(type='uniform', loc=0.0, scale=0.2)
            ...         , ScipyDistribution(type='uniform', loc=0.4, scale=0.5)])
            >>> u = f.sample()
            >>> assert u(('o1',)) <= 0.2
            >>> assert 0.4 <= u(('o2',)) <= 0.9

        Returns:

            MappingUtilityFunction
        """
        return MappingUtilityFunction(
            mapping={o: d.sample(1)[0] for o, d in self.distributions.items()}
        )

    def eval(self, offer: Outcome) -> Distribution:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:
            - You cannot return None from overriden apply() functions but raise an exception (ValueError) if it was
              not possible to calculate the Distribution.
            - Return A Distribution not a float for real-valued utilities for the benefit of inspection code.

        """
        if offer is None:
            return make_distribution(self.reserved_value)
        if self.tupelized and not isinstance(offer, tuple):
            offer = tuple(ivalues(offer))
        return self.distributions[offer]

    def xml(self, issues: list[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")

    def __str__(self):
        return pprint.pformat(self.distributions)
